#
# mkb/esxmon.py
#


#Once ESXi comes back up open the vSphere client as you need to enable OEM CIM providers.
#
#   1. Choose the Configuration tab on the host
#   2. Click Advanced Settings under the Software section
#   3. In the dialog that appears click "UserVars" on the left
#   4. Change the value of CIMOEMProvidersEnabled to 1
#   5. Click OK
#From here if you have access to the ESXi console hit  and select Restart Management Agents. Otherwise you will need to reboot the host again.

import os, string, subprocess, fnmatch, platform
import posixpath
import urllib, urllib2

from datetime import datetime, timedelta

#import paramiko
import pywbem

from archiver import *
import cron

test_class='VMware_Controller'
#test_class='CIM_Memory' # to simulate failure

default_class_list=[
    'CIM_Sensor', 
    'CIM_PowerSupply',
    'CIM_Chassis', 
    'CIM_ComputerSystem',
    'CIM_NumericSensor',
    'CIM_Memory',
    'CIM_PhysicalMemory',
    'CIM_Processor', 
    'CIM_LogRecord',
    'CIM_RecordLog',
    'CIM_EthernetPort',

    'CIM_SoftwareIdentity', 
    
    'OMC_SMASHFirmwareIdentity',
    'OMC_DiscreteSensor', 
    'OMC_Fan', 
    'OMC_PowerSupply',

    'OMC_RawIpmiSensor',
    'OMC_RawIpmiEntity',
    
    'VMware_StorageExtent',
    'VMware_Controller',
    'VMware_StorageVolume',
    'VMware_Battery',
    'VMware_SASSATAPort',
]

ExitUnknown = -1
ExitOK = 0
ExitWarning = 1
ExitCritical = 2

class ESXMon(Archiver):
    
    name='esxmon'

    types=dict( mon='mon')
    
    NS = 'root/cimv2'


    interpret=dict( HealthState = {
                        0  : ExitOK,        # Unknown
                        5  : ExitOK,        # OK
                        10 : ExitWarning,    # Degraded
                        15 : ExitWarning,    # Minor
                        20 : ExitCritical,    # Major
                        25 : ExitCritical,    # Critical
                        30 : ExitCritical,    # Non-recoverable Error
                    },
                    OperationalStatus = {
                        0  : ExitOK,            # Unknown
                        1  : ExitCritical,      # Other
                        2  : ExitOK,            # OK
                        3  : ExitWarning,       # Degraded
                        4  : ExitWarning,       # Stressed
                        5  : ExitWarning,       # Predictive Failure
                        6  : ExitCritical,      # Error
                        7  : ExitCritical,      # Non-Recoverable Error
                        8  : ExitWarning,       # Starting
                        9  : ExitWarning,       # Stopping
                        10 : ExitCritical,      # Stopped
                        11 : ExitOK,            # In Service
                        12 : ExitWarning,       # No Contact
                        13 : ExitCritical,      # Lost Communication
                        14 : ExitCritical,      # Aborted
                        15 : ExitOK,            # Dormant
                        16 : ExitCritical,      # Supporting Entity in Error
                        17 : ExitOK,            # Completed
                        18 : ExitOK,            # Power Mode
                        19 : ExitOK,            # DMTF Reserved
                        20 : ExitOK             # Vendor Reserved
                    })

    # -----------------------------------------------------------------------
    def __init__(self):
        Archiver.__init__(self)
        
        self.url=None
        self.login=None
        self.password=None

        self.class_list=[]
        self.class_include=[]
        self.class_exclude=[]
        self.property_exclude=[]
        
        self.classData = {} # Dictionary to cache class metadata
        
        self.verbose=True
        self.friendly_value=False
        
    # -----------------------------------------------------------------------
    def load(self, job, manager):
        
        log=manager.log
        now=manager.now
        
        errors, warnings, extra={}, {}, ''

        for name in ('url', 'login', 'password'):
            value=job.get(name, None)
            if not value:
                errors[name]='option mandatory'
            else:
                setattr(self, name, value)
                if name=='password':
                    self.register_password(value)

        self.verbose=boolean.get(job.get('verbose', 'yes').lower(), None)
        if self.verbose==None:
            errors['verbose']='boolean must have value in (on, yes, true, 1, off, no, false and 0)'

        self.mail_when=job.get('mail_when', 'always').lower()
        try:
            self.mail_when=int(self.mail_when)
        except ValueError:
            if not self.mail_when in ('always', 'change', 'status', 'never'):
                errors['mail_when']='must be a frequency in hours or a string in "always" or "change" or "status" or "never"'
            
        self.friendly_value=boolean.get(job.get('friendly_value', 'no').lower(), None)
        if self.friendly_value==None:
            errors['friendly_value']='boolean must have value in (on, yes, true, 1, off, no, false and 0)'

        for name in ('class_list', 'class_include', 'class_exclude', 'property_exclude'):
            value=job.get(name, None)
            try:
                lst=quoted_string_list(job.get(name, ''))
            except ValueError:
                errors[name]='not a valid quoted string list'
            else:
                setattr(self, name, lst)
                
        if not self.class_list:
            self.class_list=default_class_list
            
        class_include_warnings=[]
        for cls in self.class_include:
            if not cls in self.class_list:
                self.class_list.append(cls)
            else:
                class_include_warnings.append(cls)
                
        class_exclude_warnings=[]
        for cls in self.class_exclude:
            if cls in self.class_list:
                self.class_list.remove(cls)
            else:
                class_exclude_warnings.append(cls)
                
        if class_include_warnings:
            warnings['class_include']='classes already in the list: '+','.join(class_include_warnings)
            
        if class_exclude_warnings:
            warnings['class_exclude']='classes not in the list: '+','.join(class_exclude_warnings)
                    
        if not errors:
            # upload and setup files on the VMWARE server
            extra='class list: '+', '.join(self.class_list)
            
            # make a try 
            wbemclient=pywbem.WBEMConnection(self.url, (self.login, self.password), self.NS)
            try:
                _instance_list=wbemclient.EnumerateInstances(test_class)
            except pywbem.cim_operations.CIMError, e:
                if e.args[0]==0:
                    log.exception('maybe something wrong with URL "%s": %s', self.url, e[1])
                    errors['url']='maybe something wrong with URL "%s": %s' % (self.url, e[1])
                elif e.args[0]==pywbem.CIM_ERR_NOT_FOUND:
                    pass # The requested object could not be found.
                else:
                    errors['url']="Unknown CIM Error: code=%s, msg=%s" % (e[0], e[1])
                    log.error("Unknown CIM Error: code=%s, msg=%s", e[0], e[1])
            except pywbem.cim_http.AuthError, e: 
                log.error("Authentication error: %s", e)
                errors['login']="Authentication error: %s" % (e,)

        return errors, warnings, extra

    # -----------------------------------------------------------------------
    def friendlyValue(self, client, instance, propertyName):
       # Start out with a default empty string, in case we don't have a mapping
       mapping=None
    
       if instance.classname not in self.classData:
          # Fetch the class metadata if we don't already have it in the cache
          self.classData[instance.classname]=client.GetClass(instance.classname, IncludeQualifiers=True)
    
       # Now scan through the qualifiers to look for ValueMap/Values sets
       qualifiers=self.classData[instance.classname].properties[propertyName].qualifiers
       if 'ValueMap' in qualifiers.keys() and 'Values' in qualifiers.keys():
          vals=qualifiers['Values'].value
          valmap=qualifiers['ValueMap'].value
          value=instance[propertyName]
          # Find the matching value and convert to the friendly string
          if isinstance(value, (list, tuple)):
              mapping=[]
              for v in value:
                  for i in range(0,len(valmap)-1):
                     if str(valmap[i]) == str(v):
                         mapping.append(vals[i])
                         break
              mapping=' ('+', '.join(mapping)+')'
         
          else:
              for i in range(0,len(valmap)-1):
                 if str(valmap[i])==str(value):
                     mapping=vals[i]
                     break
       return mapping

    # -----------------------------------------------------------------------
    def run(self, command, job, manager):

        log=manager.log
        now=manager.now

        start=int(time.time())
        log.info('start url=%s', self.url)

        global_status, data, error=ExitOK, '', ''

        wbemclient=pywbem.WBEMConnection(self.url, (self.login, self.password), self.NS)

        for cls in self.class_list:
            try:
                instance_list=wbemclient.EnumerateInstances(cls)
            except pywbem.cim_operations.CIMError, e:
                if e.args[0]==pywbem.CIM_ERR_NOT_FOUND:
                    log.warning('class not found: %s (%s)', cls, e.args[1])
                    continue
                else:
                    log.exception('Unexpected error enumerating class %s', cls)
                    continue
                
            buggy=False
            for instance in instance_list:
                try:
                    elementName=instance['ElementName']
                except KeyError:
                    if not buggy:
                        log.error('Skip unknown buggy instance in class %s', cls)
                        buggy=True
                    continue
                title='===== %s %s =====' % (cls, elementName)
                for propertyName in sorted(instance.keys()):
                    fullPropertyName="%s.%s.%s" % (cls, elementName, propertyName)
                    try:
                        if propertyName in self.property_exclude or fullPropertyName in self.property_exclude or "%s.%s" % (cls, propertyName) in self.property_exclude:
                            log.debug('skip property "%s"', fullPropertyName)
                            continue
                        if not self.verbose and propertyName not in ('OperationalStatus', 'HealthState'):
                            continue
                        if instance[propertyName] is None:
                            # skip over null properties
                            continue
                        value=None
                        if self.friendly_value:
                            try:
                                value=self.friendlyValue(wbemclient, instance, propertyName)
                            except Exception, e:
                                log.exception('Retrieving friendly value for class.instance.property="%s"', fullPropertyName)
    
                        if value==None:
                            value=instance[propertyName]
                            if isinstance(value, (list, tuple)):
                                value='('+', '.join(map(str, value))+')'
                            else:
                                value=str(value)
                               
                        status=None
                        if propertyName in ('OperationalStatus', 'HealthState'):
                            status=ExitOK
                            values=instance[propertyName]
                            if not isinstance(values, (list, tuple)):
                                values=[values, ]
                            for v in values:
                                s=self.interpret[propertyName].get(v, ExitUnknown)
                                status=max(status, s)
                            if status!=ExitOK:
                                error+='%s %s %s: %s\n' % (cls, elementName, propertyName, value)
                                log.info('ERROR %s %s %s: %s', cls, elementName, propertyName, value)
                            global_status=max(global_status, status)
                        if title:
                            data+=title+'\n'
                            title=''
                            
                        if status==ExitOK:
                            status='OK ' 
                        elif status==None:
                            status='   '
                        else:
                            status='ERR'
    
                        data+='%s\t%s=%s\n' % (status, propertyName, value) 
                    except Exception, e:
                        log.exception('Unexpected error handling class.instance.property="%s"', fullPropertyName)
                        

        end=int(time.time())
        
        if global_status==ExitOK:
            backup_status='OK'
        else:
            backup_status='ERR'

        log.info('status: %s', backup_status)

        log.debug('=========== output.txt ===========')
        for line in data.split('\n'):
            log.debug('> %s',line)
        
        status=u''
        attachments=[ # the type, subtype and coding is not used
                       ('output.txt', None, data, 'text', 'plain', 'utf-8'),
                    ]
        
        if error:
            attachments.insert(0, ('error.txt', None, error, 'text', 'plain', 'utf-8'))

        status=u''
        status+='name=%s\r\n' % job['name']
        status+='program=%s\r\n' % self.name
        status+='version=%s\r\n' % self.__version__
        status+='status=%s\r\n' % backup_status
        status+='hostname=%s\r\n' % manager.hostname
        
        status+='exit_code=%d\r\n' % global_status
        status+='start=%s\r\n' % time.ctime(start)
        status+='end=%s\r\n' % time.ctime(end)
        status+='start_epoch=%d\r\n' % start
        status+='end_epoch=%d\r\n' % end

        # do I need to send an email or alert MagiKmon ?
        prev_values, magikmon, registry={}, job['magikmon'], None

        if self.mail_when in ('never', ):
            job['mail']=False # sorry for cheating
            log.info('never send any email')
        
        if not self.mail_when in ('always', 'never') or magikmon:
            try:
                import hashlib
                md5sum=hashlib.md5(data).hexdigest()
            except ImportError:
                import md5
                md5sum=md5.new(data).hexdigest()
    
            if sys.platform in ('win32', ):
                import windows
                registry=windows.Registry(job['name'])
            else:
                registry=UnixRegistry(job['name'])
                
            prev_values=registry.LoadValues()
            registry.SaveValues(dict(status=backup_status, md5sum=md5sum, last=int(time.time())))

            # check if I need to send an email
            if not self.mail_when in ('always', 'never'):
                changed_status=prev_values.get('mail.status', '<UNK>')!=backup_status
                changed_md5sum=prev_values.get('mail.md5sum', '<md5>')!=md5sum
                expired=isinstance(self.mail_when, int) and (time.time()-int(prev_values.get('mail.last', 0)))>self.mail_when
                log.debug('mail status changed=%r   report changed=%r   time expired=%r', changed_status, changed_md5sum, expired)
                if self.mail_when=='always' or expired or (isinstance(self.mail_when, int) and (changed_status or changed_md5sum)) or (self.mail_when=='change' and (changed_status or changed_md5sum)) or (self.mail_when=='status' and changed_status):
                    # send an email
                    registry.SaveValues({'mail.status':backup_status, 'mail.md5sum':md5sum, 'mail.last':int(time.time())})
                else:
                    # don't send the email ! 
                    job['mail']=False # sorry for cheating
                    log.info('don\'t send any email because the status has not changed')

            # update MagiKmon, no more then once per hour except when something changed !
            if magikmon:
                
                changed_status=prev_values.get('magikmon.status', '<UNK>')!=backup_status
                changed_md5sum=prev_values.get('magikmon.md5sum', '<md5>')!=md5sum
                expired=(time.time()-int(prev_values.get('magikmon.last', 0)))>3300
                # log.debug('magikmon status %r %r', prev_values.get('magikmon.status', '<UNK>'), backup_status)
                # log.debug('magikmon md5sum %r %r', prev_values.get('magikmon.md5sum', '<md5>'), md5sum)
                log.debug('magikmon status changed=%r   report changed=%r   time expired=%r', changed_status, changed_md5sum, expired)
                if changed_status or changed_md5sum or expired:
                    pass # send
                else:
                    log.info('don\'t update MagiKmon status this time')
                    magikmon=None

        if magikmon:
            data=urllib.urlencode(dict(status=backup_status, payload=data))
            try:
                try:
                    urllib2.urlopen(magikmon, data, timeout=30)
                except TypeError:
                    # python 2.4
                    urllib2.urlopen(magikmon, data)
            except Exception, e:
                log.exception('error updating MagiKmon status: %s', e)
            else:
                log.debug('MagiKmon updated: %s', magikmon)
                registry.SaveValues({'magikmon.status':backup_status, 'magikmon.md5sum':md5sum, 'magikmon.last':int(time.time())})
        
        if registry:
            registry.Close()
        
        return backup_status, status, attachments



#!/usr/bin/env python

import re


class CronException(Exception):
    pass

class Cron:
    
    modifier={ '1st':1, '2nd':2, '3rd':3, '4th':4, '5th':5, 'first':1, 'second':2, 'last':0 }
    
    month_lst=['jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov', 'dec']
    dayofweek_lst=['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']
    dayofmonth_lst=[str(x) for x in range(1,32)]


    unique_re=re.compile('.+')
    range_re=re.compile('([^-/]+)[-]([^-/]+)')
    simple_step_re=re.compile('([^-/]+)/([0-9]+)')
    step_re=re.compile('([^-/]+)[-]([^-/]+)/([0-9]+)')
    week_selector_re=re.compile ('w(mon|tue|wed|thu|fri|sat|sun)([0-9]*)/([0-9]+)')
    
    def __init__(self, selector):
        self.daysofmonth=set()
        self.daysofweek=set()
        self.months=set()
        self.firstdayofweek=None
        self.weekdivisor=None
        self.weekselector=None
        
        for period in selector.split(','):
            period=period.strip()
            first, last, step=None, None, '0'

            match=self.week_selector_re.match(period)
            if match:
                if self.firstdayofweek!=None:
                    raise CronException, 'week selector already set: %s' % (period,)
                self.firstdayofweek, self.weekselector, self.weekdivisor=match.groups()
                if self.weekselector:
                    self.weekselector=int(self.weekselector)
                else:
                    self.weekselector=None
                self.firstdayofweek=self.dayofweek_lst.index(self.firstdayofweek)
                self.weekdivisor=int(self.weekdivisor)
            else:
                while True:
                    match=self.step_re.match(period)
                    if match:
                        first, last, step=match.groups()
                        break
                    
                    match=self.simple_step_re.match(period)
                    if match:
                        first, step=match.groups()
                        break
    
                    match=self.range_re.match(period)
                    if match:
                        first, last=match.groups()
                        step='1'
                        break
                    
                    first=period
                    break
                
                found, size=False, 0
                for cls, pset in [(self.month_lst, self.months), (self.dayofweek_lst, self.daysofweek), (self.dayofmonth_lst, self.daysofmonth) ]:
                    try:
                        f=cls.index(first)
                    except ValueError:
                        continue
                    
                    found=pset
                    size=len(cls)
    
                    if last:
                        try:
                            l=cls.index(last)+1
                        except ValueError:
                            raise CronException, '%s not in the same range as %s' % (last, first)
                    else:
                        l=len(cls)+1
    
                    try:
                        s=int(step)
                    except ValueError:
                        raise CronException, '%s not a valid step' % (step,)
                    else:
                        if s==0:
                            l=f+1
                            s=1
                    
                if found==False:
                    raise CronException, '%s unknown' % (first,)
                
                if f>l:
                    l+=size
    
                #print period, f, l, s, range(f,l,s)
                    
                found|=set(x%size for x in range(f,l,s))
            

    def __str__(self):
        st=''
        for cls, pset in [(self.month_lst, self.months), (self.dayofweek_lst, self.daysofweek), (self.dayofmonth_lst, self.daysofmonth) ]:
            if pset:
                for i, dsp in enumerate(cls):
                    if i in pset:
                        st+=dsp+','
                        
        if self.firstdayofweek:
            if self.weekselector!=None:
                st+='(w%d-%d/%d)' % (self.firstdayofweek, self.weekselector, self.weekdivisor,)
            else:
                st+='(w%d/%d)' % (self.firstdayofweek, self.weekdivisor,)
        return st
            


if False:
    
    c=Cron('mon-fri')
    print c, '\n'
    
    c=Cron('sun-mon')
    print c, '\n'
    
    c=Cron('sun,mon-fri/2,sat')
    print c, '\n'
    
    c=Cron('jan-dec/2,thu/2,1-31/2')
    print c, '\n'
    
    c=Cron('w0/2')
    print c, '\n'
    
    c=Cron('w1/2')
    print c, '\n'

    c=Cron('sun,wsun0/2')
    print c, '\n'

    c=Cron('mon-sat,wsun0/2')
    print c, '\n'



#!/usr/bin/env python
# mksbackup -- frontent for popular Windows and Linux backup tools
# Copyright (C) Alain Spineux <alain.spineux@gmail.com>
# This program is licensed under the GNU General Public License (GPL)
#
# See http://www.magikmon.com/mksbackup for more information

import sys
import mksbackuplib

if __name__ == "__main__" and not globals().has_key('__no_execute__'):
    sys.exit(mksbackuplib.main())




#
# mkbackup/windows.py
#
# windows code


import os, sys, time, locale
import _winreg

import archiver


default_encoding=locale.getdefaultlocale()[1]
console_encoding=sys.stderr.encoding


#======================================================================
#
# Event Viewer  logging
#
#======================================================================
import win32evtlog
import win32evtlogutil
import winerror
import win32con
import pywintypes

evt_dict={
        win32con.EVENTLOG_AUDIT_FAILURE:'AUDIT_FAILURE',
        win32con.EVENTLOG_AUDIT_SUCCESS:'AUDIT_SUCCESS',
        win32con.EVENTLOG_INFORMATION_TYPE:'INF',
        win32con.EVENTLOG_WARNING_TYPE:'WAR',
        win32con.EVENTLOG_ERROR_TYPE:'ERR'
        }

# ---------------------------------------------------------------------
def FormatEv(ev_obj, logtype):
    computer=str(ev_obj.ComputerName)
    # cat=str(ev_obj.EventCategory)
    level=str(ev_obj.EventType )
    record=str(ev_obj.RecordNumber)
    evt_id=str(winerror.HRESULT_CODE(ev_obj.EventID))
    evt_type=evt_dict.get(ev_obj.EventType, 'UNK')
    msg=win32evtlogutil.SafeFormatMessage(ev_obj, logtype)
    epoch=int(ev_obj.TimeGenerated)
    msg=u'=== eventid=%d eventtype=%s epoch=%d time="%s" ===\r\n%s' % ( ev_obj.EventID, evt_type, epoch, time.ctime(epoch), msg)
    #print ev_obj.EventID, evt_type, int(ev_obj.TimeGenerated), level, msg.encode('UTF-8')
    return msg 

# ---------------------------------------------------------------------
def ReadEvLog(logtype, source, log, start, end=None):
    # If any event is not 'INFO' => return 'ERR'
    # If the last event is not an 'INFO' => return 'WAR' (in fact when not event are present)
    # Else return 'OK'
    flags = win32evtlog.EVENTLOG_BACKWARDS_READ|win32evtlog.EVENTLOG_SEQUENTIAL_READ
    try:
        hand=win32evtlog.OpenEventLog(None, logtype) # None for localhost
    except pywintypes.error, e:
        log.error('EventLog error %r', e)
        return 'ERR', u'EventLog error %r' % (e, )
    else:
        cont, output, status=True, u'', ''
        while cont:
            events=win32evtlog.ReadEventLog(hand,flags,0)
            for ev_obj in events:
                if str(ev_obj.SourceName)!=source:
                    continue
                if int(ev_obj.TimeGenerated)<start:
                    cont=False
                    break
                if ev_obj.EventType!=win32con.EVENTLOG_INFORMATION_TYPE:
                    status='ERR'
    
                if ev_obj.EventID==8019 and ev_obj.EventType==win32con.EVENTLOG_INFORMATION_TYPE and status!='ERR':
                    status='OK'
                        
                if output:
                    output=FormatEv(ev_obj, logtype)+u'\r\n'+output
                else:
                    output=FormatEv(ev_obj, logtype)
    
            cont=cont and events
        win32evtlog.CloseEventLog(hand)
        
    if status=='':
        status='ERR'

    return status, output

# ---------------------------------------------------------------------
def WindowsErrorDecode(exception, encoding=default_encoding):
    """convert WindowsError or OSError exception into unicode string, 
    some message can contains non ascii chars in non US windows version"""
    
    # if isinstance(exception, (WindowsError, OSError))

    st=str(exception)
    try:
        return st.decode(encoding)
    except UnicodeDecodeError:
        return st.decode('ascii', 'replace')

# ---------------------------------------------------------------------------
def windows_list_dir(target_dir, log, encoding=default_encoding):
    try:
        dir_out, total_size=archiver._list_dir(target_dir, log) 
    except Exception, e:
        log.error('listing target directory: %s', WindowsErrorDecode(e, encoding))
        dir_out='error listing directory "%s": %s\r\n' % (target_dir, WindowsErrorDecode(e, encoding))
        total_size=0
        
    return dir_out, total_size


# =====================================================================
class Registry:

    sub_key="SOFTWARE\\MagiKSys\\MKSBackup\\%s"

    def __init__(self, master_key):
        self.home_key=self.sub_key % (master_key, )
        self.reg=None
        try:
            self.reg=_winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, self.home_key, 0, _winreg.KEY_READ|_winreg.KEY_WRITE)
        except EnvironmentError:
            self.reg=_winreg.CreateKey(_winreg.HKEY_LOCAL_MACHINE, self.home_key)
            _winreg.CloseKey(self.reg)
            self.reg=_winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, self.home_key, 0, _winreg.KEY_READ|_winreg.KEY_WRITE)
        
    def LoadValues(self, keys=[]):
        values=dict()
        try:
            i=0
            while True:
                k, v, _t=_winreg.EnumValue(self.reg, i)
                i+=1
                if not keys or k in keys:
                    values[k]=v
        except WindowsError:
            pass
        
        return values    
                
    def SaveValues(self, values):
        for k, v in values.iteritems():
            _winreg.SetValueEx(self.reg, k, None, _winreg.REG_SZ, str(v))

    def Close(self):
        if self.reg:
            _winreg.CloseKey(self.reg)




#
# (C) Copyright 2003, 2004 Hewlett-Packard Development Company, L.P.
# (C) Copyright 2006, 2007 Novell, Inc. 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Tim Potter <tpot@hp.com>
# Author: Bart Whiteley <bwhiteley@suse.de>

"""
Subclasses of builtin Python types to remember CIM types.  This is
necessary as we need to remember whether an integer property is a
uint8, uint16, uint32 etc, while still being able to treat it as a
integer.
"""

from datetime import tzinfo, datetime, timedelta
import re

class MinutesFromUTC(tzinfo):
    """Fixed offset in minutes from UTC."""
    def __init__(self, offset):
        self.__offset = timedelta(minutes = offset)
    def utcoffset(self, dt):
        return self.__offset
    def dst(self, dt):
        return timedelta(0)

class CIMType(object):
    """Base type for all CIM types."""
    
class CIMDateTime(CIMType) :
    """A CIM DateTime."""

    def __init__(self, dtarg):
        """Construct a new CIMDateTime

        Arguments:
        dtarg -- Can be a string in CIM datetime format, a datetime.datetime, 
            or a datetime.timedelta. 

        """

        self.cimtype = 'datetime'
        self.__timedelta = None
        self.__datetime = None
        if isinstance(dtarg, basestring):
            date_pattern = re.compile(r'^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})\.(\d{6})([+|-])(\d{3})')
            s = date_pattern.search(dtarg)
            if s is not None:
                g = s.groups()
                offset = int(g[8])
                if g[7] == '-':
                    offset = -offset
                self.__datetime = datetime(int(g[0]), int(g[1]), 
                                           int(g[2]), int(g[3]), 
                                           int(g[4]), int(g[5]), 
                                           int(g[6]), MinutesFromUTC(offset))
            else:
                tv_pattern = re.compile(r'^(\d{8})(\d{2})(\d{2})(\d{2})\.(\d{6})(:)(000)')
                s = tv_pattern.search(dtarg)
                if s is None:
                    raise ValueError('Invalid Datetime format "%s"' % dtarg)
                else:
                    g = s.groups()
                    self.__timedelta =  timedelta(days=int(g[0]),
                                                  hours=int(g[1]),
                                                  minutes=int(g[2]),
                                                  seconds=int(g[3]),
                                                  microseconds=int(g[4]))
        elif isinstance(dtarg, datetime):
            self.__datetime = dtarg; 
        elif isinstance(dtarg, timedelta):
            self.__timedelta = dtarg
        elif isinstance(dtarg, CIMDateTime):
            self.__datetime = dtarg.__datetime
            self.__timedelta = dtarg.__timedelta
        else:
            raise ValueError('Expected datetime, timedelta, or string')

    @property
    def minutes_from_utc(self):
        """Return the timezone as +/- minutes from UTC"""
        offset = 0
        if self.__datetime is not None and \
                self.__datetime.utcoffset() is not None:
            offset = self.__datetime.utcoffset().seconds / 60
            if self.__datetime.utcoffset().days == -1:
                offset = -(60*24 - offset)
        return offset

    @property
    def datetime(self):
        return self.__datetime

    @property
    def timedelta(self):
        return self.__timedelta

    @property
    def is_interval(self):
        return self.__timedelta is not None

    @staticmethod
    def get_local_utcoffset():
        """Return minutes +/- UTC for the local timezone"""
        utc = datetime.utcnow()
        local = datetime.now()
        if local < utc:
            return - int(float((utc - local).seconds) / 60 + .5)
        else:
            return int(float((local - utc).seconds) / 60 + .5)

    @classmethod
    def now(cls, tzi=None):
        if tzi is None:
            tzi = MinutesFromUTC(cls.get_local_utcoffset())
        return cls(datetime.now(tzi))

    @classmethod
    def fromtimestamp(cls, ts, tzi=None):
        if tzi is None:
            tzi = MinutesFromUTC(cls.get_local_utcoffset())
        return cls(datetime.fromtimestamp(ts, tzi))

    def __str__ (self):
        if self.is_interval:
            hour = self.__timedelta.seconds / 3600
            minute = (self.__timedelta.seconds - hour * 3600) / 60
            second = self.__timedelta.seconds - hour * 3600 - minute * 60
            return '%08d%02d%02d%02d.%06d:000' % \
                    (self.__timedelta.days, hour, minute, second, 
                            self.__timedelta.microseconds)
        else:
            offset = self.minutes_from_utc
            sign = '+'
            if offset < 0:
                sign = '-'
                offset = -offset
            return '%d%02d%02d%02d%02d%02d.%06d%s%03d' % \
                   (self.__datetime.year, self.__datetime.month, 
                    self.__datetime.day, self.__datetime.hour,
                    self.__datetime.minute, self.__datetime.second, 
                    self.__datetime.microsecond, sign, offset)

    def __repr__ (self):
        return '%s(%s)'%(self.__class__.__name__, `str(self)`)

    def __getstate__(self):
        return str(self)

    def __setstate__(self, arg):
        self.__init__(arg)

    def __cmp__(self, other):
        if self is other:
            return 0
        elif not isinstance(other, CIMDateTime):
            return 1
        return (cmp(self.__datetime, other.__datetime) or 
                cmp(self.__timedelta, other.__timedelta))


# CIM integer types

class CIMInt(CIMType, long):
    pass

class Uint8(CIMInt):
    cimtype = 'uint8'

class Sint8(CIMInt):
    cimtype = 'sint8'

class Uint16(CIMInt):
    cimtype = 'uint16'

class Sint16(CIMInt):
    cimtype = 'sint16'

class Uint32(CIMInt):
    cimtype = 'uint32'

class Sint32(CIMInt):
    cimtype = 'sint32'

class Uint64(CIMInt):
    cimtype = 'uint64'

class Sint64(CIMInt):
    cimtype = 'sint64'

# CIM float types

class CIMFloat(CIMType, float):
    pass

class Real32(CIMFloat):
    cimtype = 'real32'

class Real64(CIMFloat):
    cimtype = 'real64'

def cimtype(obj):
    """Return the CIM type name of an object as a string.  For a list, the
    type is the type of the first element as CIM arrays must be
    homogeneous."""
    
    if isinstance(obj, CIMType):
        return obj.cimtype

    if isinstance(obj, bool):
        return 'boolean'

    if isinstance(obj, (str, unicode)):
        return 'string'

    if isinstance(obj, list):
        return cimtype(obj[0])

    if isinstance(obj, (datetime, timedelta)):
        return 'datetime'

    raise TypeError("Invalid CIM type for %s" % str(obj))



def atomic_to_cim_xml(obj):
    """Convert an atomic type to CIM external form"""
    if isinstance(obj, bool):
        if obj:
            return "true"
        else:
            return "false"

    elif isinstance(obj, CIMDateTime):
        return unicode(obj)

    elif isinstance(obj, datetime):
        return unicode(CIMDateTime(obj))

    elif obj is None:
        return obj
    elif cimtype(obj) == 'real32':
        return u'%.8E' % obj
    elif cimtype(obj) == 'real64':
        return u'%.16E' % obj
    else:
        return unicode(obj)




# /bin/env python
# setup.py
# (c) alain.spineux@gmail.com

import sys, os

if 'bdist_egg' in sys.argv and 'py2exe' in sys.argv :
    # both "patch" distutil and py2exe cannot be mix together 
    print "don't mix bdist_egg and py2exe in the same command !"
    sys.exit(1)

if 'bdist_egg' in sys.argv:
    from setuptools import setup
else:
    from distutils.core import setup

if 'py2exe' in sys.argv:
    # ... http://www.py2exe.org/index.cgi/win32com.shellhttp://www.py2exe.org/index.cgi/win32com.shell
    # ModuleFinder can't handle runtime changes to __path__, but win32com uses them

    # py2exe 0.6.4 introduced a replacement modulefinder.
    # This means we have to add package paths there, not to the built-in
    # one.  If this new modulefinder gets integrated into Python, then
    # we might be able to revert this some day.
    # if this doesn't work, try import modulefinder
    try:
        import py2exe.mf as modulefinder
    except ImportError:
        import modulefinder
    import win32com
    for p in win32com.__path__[1:]:
        modulefinder.AddPackagePath("win32com", p)
    for extra in ["win32com.shell"]: #,"win32com.mapi"
        __import__(extra)
        m = sys.modules[extra]
        for p in m.__path__[1:]:
            modulefinder.AddPackagePath(extra, p)

    from py2exe.build_exe import py2exe

    class build_zip(py2exe):
        """This class inherit from py2exe, builds the exe file(s), then creates a ZIP file."""
        def run(self):
            
            import zipfile
            # initialize variables and create appropriate directories in 'buid' directory
            # please don't change 'dist_dir' in setup()
             
            orig_dist_dir=self.dist_dir
            self.mkpath(orig_dist_dir)
            zip_filename=os.path.join(orig_dist_dir, '%s-%s-win32.zip' % (self.distribution.metadata.name, self.distribution.metadata.version,))
            bdist_base=self.get_finalized_command('bdist').bdist_base
            dist_dir=os.path.join(bdist_base, '%s-%s' % (self.distribution.metadata.name, self.distribution.metadata.version, ))
            self.dist_dir=dist_dir
            print 'dist_dir is %s' % (dist_dir, )

            # let py2exe do it's work.
            py2exe.run(self)
            
            # remove zipfile if exists
            if os.path.exists(zip_filename):
                os.unlink(zip_filename)
            
            # create the zipfile
            print 'Building zip file %s' % (zip_filename, )
            zfile=zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED)
            for root, dirs, files in os.walk(dist_dir):
                for f in files:
                    filename=os.path.join(root, f)
                    zfile.write(filename, os.path.relpath(filename, bdist_base))
                    
#            zfile.writestr('EGG-INFO/PKG-INFO', 'This file is required by PyPI to allow upload as and bdist_egg archive')
            zfile.close()
            # pip and easy_install try to use the win32 .zip archive as a source file
            # whatever it's type is and faile ! This file cannot be uploaded on PyPI
            # self.distribution.dist_files.append(('bdist_dumb', '2.3', zip_filename))


basename='mksbackup'
from mksbackuplib.version import __version__ as version
print 'VERSION', version

extra_options = {}
doc_dir='share/doc/%s-%s' % (basename, version)

cmdclass = {}

if 'py2exe' in sys.argv and os.name=='nt':
    doc_dir='doc'
    cmdclass = {"py2exe": build_zip}
    win32zip='%s-%s' % (basename, version)
    py2exe_options =  { # 'ascii': True, # exclude encodings
                        'packages':['email', 'new'], 
                        'dll_excludes': ['w9xpopen.exe'], # no support for W98 
                        'compressed':True, 
                        # 'dist_dir': ????  # !!! DONT CHANGE distdir HERE 
                        # exclude big and unused python packages from the binary
                        'excludes': [ 'difflib', 'doctest', 'pdb'],
                        }
    extra_options = { 'console': ['mksbackup', ],  # list of scripts to convert into console exes
                      'windows': [],          # list of scripts to convert into gui exes
                      'options': { 'py2exe': py2exe_options, } ,
                    }
    if '--single-file' in sys.argv[1:]:
        sys.argv.remove('--single-file')
        py2exe_options.update({ 'bundle_files': 1, })
        extra_options.update({ 'zipfile': None, }) # don't build a separate zip file with all libraries, put them all in the .exe  

data_files=[ (doc_dir, [ 'README.txt', 'Changelog.txt', 'LICENSE.txt', 'ghettoVCB.txt', 'sample.ini' ]),
#             ('', ['__main__.py', ]),
           ]

# if 'bdist_egg' in sys.argv:
#     data_files.append(('', ['__main__.py', ]))

setup(name='mksbackup',
      version=version, 
      author='Alain Spineux',
      author_email='alain.spineux@gmail.com',
      url='http://www.magikmon.com/mksbackup',
      description='Front-end for popular Windows, Linux and VMware backup tools. Sending a mail report at end of backup.',
      long_description='MKSBackup can run Windows ntbackup, wbadmin or Linux tar or even VMWare ESX(i) '
                       'ghettoVCB.sh using settings found in a .ini file. When done MKSBackup can '
                       'send emails including a full report. The email subject display an unambiguous '
                       'status "ERR" or "OK" to allow easy filtering in any mail reader. '
                       'MKSBackup handle custom target destination directory and filename naming based '
                       'on date and time. '
                       'MKSBackup can also monitor VMWare ESX(i) hardware via included esxmon module. '
                       'Backup reports can also be sent to MagiKmon http://www.magikmon.com to '
                       'be parsed and compiled into more synthetic reports and graphs. MagiKmon can '
                       'generate alerts when backups fails for too long. '
                       'MKSBackup is the new name for old MKBackup.'
                       'A .exe for windows and a runnable .egg that don''t need installation are '
                       'available from the site at http://www.magikmon.com/mksbackup'
                       ,
      license='GPL',
      py_modules = [ '__main__', ],
      packages=[ 'mksbackuplib', 'mksbackuplib.pywbem'],
      scripts=['mksbackup',],

      data_files=data_files,

      classifiers=["Development Status :: 5 - Production/Stable",
                  "Environment :: Console",
                  "License :: OSI Approved :: GNU General Public License (GPL)",
                  "Topic :: System :: Archiving :: Backup",
                  "Operating System :: OS Independent",
                  ],

      cmdclass = cmdclass,

      **extra_options)





#!/usr/bin/python

import SocketServer
import sys, os, socket

class FTPsession(SocketServer.BaseRequestHandler):
    """One FTP Session from login to exit"""
    
    block_size=65535
    
    def setup(self):
        self.data_socket=None
        try:
            self.options=self.server.options
        except:
            self.options={}

        self.root=os.path.abspath(self.options.get('root', os.getcwd()))
        self.user=self.options.get('user', None)
        self.password=self.options.get('password', None)
        self.debug=self.options.get('debug', None)
        # no password

    def abs_path(self, path):
        path=os.path.abspath(os.join(self.root_dir, path))

    def response(self, st):
        """Send a response to the client"""
        if self.debug:
            print 'response', st
        self.request.send("%s\r\n" % (st,) )

    def cmd_quit(self, args):
        self.response("221 Have a good one!")
        raise SystemExit
    
    def cmd_user(self, args):
        if args[0] != self.user:
            self.response("502 Invalid user")
        else:
            self.response("331 User name ok")
    
    def cmd_pass(self, args):
        if self.password and args[0]!=self.password:
            self.response("530 Not logged in")
        else:
            self.response("230 User logged in")
    
    def cmd_dummy(self, args):
        self.response("200 OK ignored")
    
    cmd_rein=cmd_dummy
    cmd_type=cmd_dummy
    cmd_noop=cmd_dummy
    cmd_cwd=cmd_dummy
    cmd_allo=cmd_dummy
    
    def cmd_syst(self, args):
        self.response("215 MKSBackup ftp")

    def cmd_port(self, args):
        try:
            parts=args[0].split(',')
            if self.data_socket:
                self.data_socket.close()
                self.datas_ocket=None
            port=int(parts[4])*256 + int(parts[5])
            host='.'.join(parts[0:4])
            self.data_socket=socket(socket.AF_INET, socket.SOCK_STREAM)
            self.data_socket.connect((host, port))
        except IndexError:
            self.response('500 port error')
            raise
        else:
            self.response('200 %s:%d' % (host, port))

    def cmd_pasv(self, args):
        if self.data_socket:
            self.data_socket.close()
            self.datas_ocket=None
        
        local_ip=self.request.getsockname()[0]
        
        self.data_socket=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.data_socket.bind((local_ip, 0))
        
        self.data_socket.listen(1) 
        
        port=self.data_socket.getsockname()[1]
        self.response('227 Entering passive mode (%s,%d,%d).' % (
                                local_ip.replace('.', ','), port / 256, port % 256))

        conn, addr=self.data_socket.accept()
        self.data_socket.close()
        self.data_socket=conn

    def cmd_list(self, args):
        if args:
            dirname=args[0]
        else:
            dirname="."
        # should check if can access the dirname
        lst = os.listdir(self.root)
        
        if lst is not None:
            self.response("150 opening ASCII connection for file list")
            data="\r\n".join(lst+[''])
            self.data_socket.send(data)
            self.data_socket.close()
            self.response("226 transfer complete")
        else:
            self.response("553 outside base directory of server")

    def cmd_size(self, args):
        self.response("550 not implemented")
    
    def cmd_cdup(self, args):
        self.response("553 already at the root dir")
    
    def cmd_mkd(self, args):
        target=os.path.join(self.root, args[0])
        if not os.path.isdir(target):
            try:
                os.mkdir(target)
            except OSError, e:
                self.response('521 "%s" %s' % (args[0], e,))
            else:
                self.response('257 "%s" directory created' % (args[0], ))
        else:
            self.response('521 "%s" directory already exists' % (args[0], ))
            
    def cmd_retr(self, args):
        if not args:
            self.response("500 retr syntax error")

        filename=args[0]
        if not os.path.isfile(filename):
            self.response("553 filename not found")

        filein=open(filename, 'rb')
        data=filein.read(self.block_size)
        self.response("125 Transfer started")
        while data:
            self.data_socket.send(data)
            data=filein.read(self.block_size)

        self.response("226 Closing data connection")
        self.data_socket.close()
        self.data_socket = None

    def cmd_stor(self, args):
        if not args:
            self.response("500 stor syntax error")
            return

        filename=os.path.abspath(os.path.join(self.root, args[0].lstrip('/')))
        if os.path.isdir(filename):
            self.response("553 target file is a directory")
            return

        if not filename.startswith(self.root):
            self.response("553 File name not allowed")
            return
            
        dirname=os.path.dirname(filename)
        if not os.path.isdir(dirname):
            try:
                os.makedirs(dirname)
            except Exception:
                self.response("553 directory don't exists")

            
        #mode = ['w','wb'][self.binary]
        try:
            fileout=open(filename, 'wb')
            self.response("125 Data transfer starting")
            try:
                while 1:
                    data=self.data_socket.recv(self.block_size)
                    if not data: 
                        break
                    fileout.write(data)
            except:
                raise
                pass
            fileout.close()
            self.response("226 Transfer complete.")
        except:
            self.response("451 local error")
            raise

    def cmd_dele(self, args):
        self.response("502 DELE not implemented")
            
    def handle(self):
        self.peername = self.request.getpeername()
        self.response("220 MKSBackup ftp ready")
        while 1:
            cmd_line=''
            while cmd_line.find('\n')==-1:
                data=self.request.recv(1024)
                if not data:
                    break
                cmd_line+=data
                
            if self.debug:
                print 'request from', self.peername, cmd_line

            if (not cmd_line): 
                break
            cmd_line=cmd_line.rstrip()
            cmd_args=cmd_line.split(' ', 1) # !!! handle only one arg !
            cmd=cmd_args[0].lower()
            
            if cmd == "quit":
                self.response("221 OK, bye")
                break

            try:
                method=getattr(self, 'cmd_'+cmd)
            except AttributeError:
                self.response("502 %s not implemented" % (cmd, ))
            else:
                try:
                    method(cmd_args[1:])
                except:
                    self.response("451 local error")
                    raise

class FTPserver(SocketServer.ThreadingTCPServer):    
    
    allow_reuse_address = True
    daemon_threads = True
    
    def __init__(self, addr=('',0), handler=FTPsession, **options):
        self.options = options
        allowed=self.options.get('allowed','')
        if allowed:
            self.allowed=map(socket.gethostbyname, allowed.split(','))
        else:
            self.allowed=[]
                   
        SocketServer.ThreadingTCPServer.__init__(self, addr, handler)

    def verify_request(self, reqsocket, client_address):
        return not self.allowed or client_address[0] in self.allowed

def ftpmain(argv):
    from optparse import OptionParser

    parser=parser=OptionParser(version='%%prog')
    parser.set_usage('%prog [options]\n\n')
    
    parser.add_option("-P", "--port", dest="port", default='21', help="listen on port", metavar="PORT")
    parser.add_option("-r", "--root", dest="root", default='/', help="root directory", metavar="DIR")
    parser.add_option("-u", "--user", dest="user", default='ftp', help="unique user", metavar="USER")
    parser.add_option("-p", "--password", dest="password", default=None, help="user password", metavar="PASSWORD")
    parser.add_option("-1", "--once", dest="once", action="store_true", default=False, help="serve only one request")
    parser.add_option("-a", "--allowed", dest="allowed", default='', help="comma separated list of host allowed to connect", metavar="HOST_LIST")
    parser.add_option("-b", "--bind", dest="bind", default='', help="bind to a specific address", metavar="address")
    
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False, help="display debugging")
    
    cmd_options, cmd_args=parser.parse_args(argv)

    server=FTPserver(addr=('',int(cmd_options.port)), handler=FTPsession, root=cmd_options.root, user=cmd_options.user, password=cmd_options.password, allowed=cmd_options.allowed, debug=cmd_options.debug)

    print 'ftp server listening on', server.socket.getsockname()
    print 'user is "%s"' % (cmd_options.user, )
    if not cmd_options.password:
        print 'password is <empty>'
    else:
        print 'password is "%s"' % (cmd_options.password, )

    if cmd_options.once:
        server.handle_request()
    else:
        server.serve_forever()
    
if __name__ == '__main__':

    ftpmain(sys.argv)



#!/bin/env python
#
# main.py
#
# (c) Alain Spineux alain.spineux@gmail.com
#
# mksbackup is a front-end for popular and free linux and windows archiver tools.
#
# Supported archiver:
# - ntbackup
# - wbadmin
# - tar
# - ghettoVCB

# Features:
# - job definitions are stored in a configuration file
# - the destination of the archive can be set to different location depending
#   on the day, the week or the month to allow very complex schema of backup
# - send and email at the end of job
# - the email contains all information to evalute the reliability of the backup jobs
#
#
# mksbackup is released under the GNU GPL license
#

import sys, os, subprocess, time, smtplib, calendar, re, socket, urllib, urllib2, logging, logging.handlers, codecs
import traceback, StringIO 
import platform, locale 
from optparse import OptionParser
import ConfigParser

from datetime import datetime, timedelta

from archiver import boolean, check_mail_config, Destinations, sendmail, Manager, send_mail_report, ftp_url_re

import cron

if sys.platform in ('win32', ):
    from windows import WindowsErrorDecode as ErrorDecode
else:
    from archiver import ErrorDecode

from version import __version__

last_version_url='http://www.magikmon.com/download/mksbackup/last2.txt'

def repl_ftp_password(m):
    """used by re.sub to replace ftp password by ***********"""
    username, password, host, port, directory=m.group('username', 'password', 'host', 'port', 'directory')
    st='ftp://'
    if username:
        host='@'+host
        st+=username
    if password:
        st+=':********'
        
    st+=host
    if port:
        st+=':'+port
    if directory:
        st+=directory
    return st
        
        
def gen_archiver(name):
    name=name.lower()
    if name=='ntbackup':
        import ntbackup
        return ntbackup.NTBackup
    elif name=='tar':
        import tar
        return tar.Tar
    elif name=='ghettovcb':
        import ghettovcb
        return ghettovcb.Ghettovcb
    elif name=='esxmon':
        import esxmon
        return esxmon.ESXMon
    elif name=='wbadmin':
        import wbadmin
        return wbadmin.Wbadmin
    elif name=='wbadminsys':
        import wbadmin
        return wbadmin.WbadminSys
    else:
        return None

# ---------------------------------------------------------------------------
def update_check(version, backend):
    """Check if and new version of MKSBackup is available and return the
    announcement message if available
    """
    
    up2date, msg='yes', ''
    data=urllib.urlencode(dict(version=version, backend=backend, platform=sys.platform, osver=platform.version()[:10]))
    current=__version__.split('.')

    try:
        data=urllib2.urlopen('%s?%s' % (last_version_url, data)).read()
#        data=urllib2.urlopen(last_version_url, data).read()
        state='version'
        msg=''
        vall=vbackend=None
        for line in data.split('\n'):
            line=line.strip()
            if not line:
                state='msg'
                continue
            if state=='version':
                try:
                    prog, ver=line.split(':')
                except ValueError:
                    continue
                else:
                    prog=prog.strip()
                    ver=ver.strip()
                    if prog==backend:
                        vbackend=ver
                    elif prog=='*':
                        vall=ver
            elif state=='msg':
                msg+=line+'\n'

        if vbackend==None:
            vbackend=vall
            
        if vbackend!=None:
                    
            last=vbackend.split('.')
            for i in range(min(len(current), len(last))):
                if int(current[i])<int(last[i]):
                    up2date='no'
                    break
            
    except (urllib2.URLError, urllib2.HTTPError):
#        raise
        up2date='unk'
    except Exception, e:
#        raise
        up2date='unk'

    return up2date, msg

# ---------------------------------------------------------------------------        
def gen_status(start, job, status, end):

    if end==None:
        end=int(time.time())
    st=u''
    st+='name=%s\r\n' % job['name']
    st+='program=%s\r\n' % job.get('program','')
    st+='version=%s\r\n' % __version__
    st+='status=%s\r\n' % (status, )
    st+='hostname=%s\r\n' % platform.node()
    st+='start=%s\r\n' % time.ctime(start)
    st+='end=%s\r\n' % time.ctime(end)
    st+='start_epoch=%d\r\n' % start
    st+='end_epoch=%d\r\n' % end
    return st

# ---------------------------------------------------------------------------        
def mail_exception(start, job, extra, manager, subject):
    
    status=gen_status(start, job, 'ERR', None)
    status+='runtime_error=%r\r\n' % sys.exc_info()[1]
    
    msg_body=job['msg_header']
    msg_body+='\nUnexpected error, read "traceback.txt" for more\n\n'
    msg_body+='\n%s\n' % status
    msg_body+=extra+'\n'

    try:
        tb=StringIO.StringIO()
        exctype, value = sys.exc_info()[:2]
        traceback.print_exc(None, tb)
        tb_st=tb.getvalue()
        tb.close()
    except Exception, e:
        tb_st='Traceback failed: %r' % e
    
    attachments=manager.attachments[:]
    attachments+=[ # the type, subtype and coding is not used
           ('traceback.txt', None, tb_st.decode('ascii', 'replace'), 'text', 'plain', 'utf-8'), # ascii is already utf8, no need to encode
           ('status.txt', None, status.encode('utf-8'), 'text', 'plain', 'utf-8'),
    ]

    try:
        log_output=manager.stop_logging()
        attachments.append(('logging.txt', None, log_output.encode('utf-8'), 'text', 'plain', 'utf-8'))
    except Exception:
        msg_body+='\nminor error when attaching logging\n'                      
    
    errmsg=sendmail(job, subject, msg_body, attachments, manager.log)


class MyFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, encoding='utf-8'):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.encoding=encoding
 
    def format(self, record):
        t=logging.Formatter.format(self, record)
        return t
        # ASX
        if isinstance(t, unicode):
            t=t.encode(self.encoding, 'replace')
        return t

#======================================================================
#
# Main
#
#======================================================================

#
# Parse the command line
#
def main():

    if len(sys.argv)>1 and sys.argv[1]=='ftp':
        args=sys.argv[:]
        args.remove('ftp')
        import ftpd
        ftpd.ftpmain(args)
        sys.exit(0)
    
    parser=parser=OptionParser(version='%%prog %s' % __version__ )
    parser.set_usage('%prog [options] command [job..]\n\n'
                     '\tcommand in "backup", "check", "checkmail"\n\n'
                     '\tuse "ftp" in first place to start buil-in ftp server,\n'
                     '\tuse "ftp -h" in this order for more')
    
    parser.add_option("-c", "--config", dest="config", default='mksbackup.ini', help="use another ini file", metavar="configfile")
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False, help="switch to debug level, increase verbosity")
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true", default=True, help="write logging to the terminal (default)")
    parser.add_option("-q", "--quiet", dest="verbose", action="store_false", default=True, help="don't write logging to the terminal")
    parser.add_option("-l", "--logfile", dest="logfile", default='mksbackup.log', help="log to this file", metavar="logfile")
    parser.add_option("-e", "--encoding", dest="encoding", default='auto', help="configuration file encoding", metavar="encoding")
    
    parser.add_option("-i", "--install", dest="install", action="store_true", help="install and pre-setup a task in scheduler")
    
    cmd_options, cmd_args=parser.parse_args(sys.argv)
    
    if cmd_options.install or len(sys.argv)==1:
        import install
        try:
            res=install.Install()
        except Exception, e:
            print 'Unexpected Error: %s' % (e, )
            raise
        sys.exit(res)
    
    root_handler=logging.handlers.RotatingFileHandler(cmd_options.logfile, 'a', 1024**2, 2)
    root_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)-3.3s %(message)s', '%Y-%m-%d %H:%M:%S'))
    log=logging.getLogger()
    log.setLevel(logging.INFO)
    log.addHandler(root_handler)
    
    if cmd_options.debug:
        log.setLevel(logging.DEBUG)
    else:
        # this stop early the logging process of loglevel<=DEBUG and should 
        # improve performance a little bit
        logging.disable(logging.DEBUG)
    
    logging.getLogger("paramiko").setLevel(logging.WARNING)
        
    if cmd_options.verbose:
        console=logging.StreamHandler()
        encoding=sys.stderr.encoding
        if encoding==None:
            encoding='ascii'
        console.setFormatter(MyFormatter('%(asctime)s,%(msecs)03d %(levelname)-3.3s %(message)s', '%H:%M:%S', encoding=encoding))
        log.addHandler(console)
    
    log.info(u'start version=%s cmd=%r', __version__, sys.argv)
    
    if len(cmd_args)<2:
        parser.error('no "command" set')
        
    command=cmd_args[1]
    if command.lower() not in ('backup', 'check', 'checkmail'):
        parser.error('invalid command "%s"' % command)
    
    command=command.lower()
    
    if command=='backup':
        command_display='BACKUP'
    else:
        command_display='CONFIG'
        
    job_list=cmd_args[2:]
    
    if not job_list:
        parser.error('no "job"')
    
    # define a default_encoding
    default_encoding=locale.getdefaultlocale()[1]
    console_encoding=sys.stderr.encoding
    
    if not default_encoding:
        # use default per platform
        if sys.platform in ('win32', ):
            default_encoding='windows-1252'
        else:
            default_encoding='utf-8'
    
    if not console_encoding:
        # use default per platform
        if sys.platform in ('win32', ):
            console_encoding='cp850'
        else:
            console_encoding='utf-8'

    #
    # load the configuration file
    #
    config_default=dict(verify='no',
                        update_check='yes',
                        smtp_host='127.0.0.1',
                        smtp_mode='normal',
                        night_shift='yes',
                        attachment_size='1000',
                        attachment_gzip='100',
                        )
    
    try:
        config_file=open(cmd_options.config, 'r')
    except IOError, e:
        log.error('error reading configuration file "%s": %s', cmd_options.config, ErrorDecode(e, default_encoding))
        sys.exit(1)
    
    config_text=config_file.read()
    config_file.close()
    config_file=None
    
    if cmd_options.encoding=='auto':
        # try to guess the file encoding  
        bomdict = { codecs.BOM_UTF8 : ('utf-8', 1),  # or ( 'utf_8_sig', 0) but utf_8_sig dont exist in 2.4
                    codecs.BOM_UTF16_BE : ('utf-16', 0),
                    codecs.BOM_UTF16_LE : ('utf-16', 0) }
    
        for bom, (encoding, skip) in bomdict.items():
            if config_text.startswith(bom):
                cmd_options.encoding=encoding
                cmd_options.skip=skip
                config_file=codecs.open(cmd_options.config, 'r', encoding)
                if skip:
                    config_file.read(skip)
                break
        
        if cmd_options.encoding=='auto':
            # if no BOM use default
            cmd_options.encoding=default_encoding
            config_file=codecs.open(cmd_options.config, 'r', cmd_options.encoding)
    else:
        try:
            config_file=codecs.open(cmd_options.config, 'r', cmd_options.encoding)        
        except LookupError, e:
            log.error('error reading configuration file %s: %s', cmd_options.config, ErrorDecode(e, default_encoding))
            sys.exit(1)
    
    config=ConfigParser.RawConfigParser(config_default)
    try:
        config.readfp(config_file)
    except (IOError, UnicodeDecodeError), e:
        log.error('error reading configuration file %s: %s', cmd_options.config, ErrorDecode(e, default_encoding))
        sys.exit(1)
    
    err, failed_job=0, job_list[:]
    for job_name in job_list:
        try:
            program=config.get(job_name, 'program')
        except ConfigParser.NoOptionError, e:
            log.error('no "program" set for job "%s"', job_name)
            err+=1
        except ConfigParser.NoSectionError, e:
            log.error('job not found: %s',job_name)
            err+=1
        else:
            if not gen_archiver(program):
                log.error('program "%s" unknown in job "%s"', program, job_name)
                err+=1
            else:
                failed_job.remove(job_name)
                
                # check for update for this back-end
                
                up2date, up2date_msg='unk', ''
                check_for_update=config.get(job_name, 'update_check')
                if boolean.get(check_for_update.lower(), None)!=False:
                    up2date, up2date_msg=update_check(__version__, program)
                    if up2date=='no':
                        log.warning('!'*60)
                        for line in up2date_msg.split('\n'):
                            if line:
                                log.warning(line)
                        log.warning('!'*60)

                config.set(job_name, 'up2date', up2date)
                config.set(job_name, 'up2date_msg', up2date_msg)
    
    if err>0:
        parser.error('in jobs: %s' % ' '.join(failed_job))
        sys.exit(3)
    
    exit_code=0
    
    # remove password from .INI file 
    sanitized_config_file_content=codecs.open(cmd_options.config, 'r', cmd_options.encoding).read()
    for password_key in ('password', 'smtp_password',):
        sanitized_config_file_content=re.sub('(?m)^#?(?P<key>\s*%s\s*)=(?P<value>.*)$' % (password_key,), '\g<key>=********', sanitized_config_file_content)

    sanitized_config_file_content=ftp_url_re.sub(repl_ftp_password, sanitized_config_file_content)
    
    manager=Manager(platform.node(), log)
    manager.default_encoding=default_encoding
    manager.console_encoding=console_encoding
    manager.formater=MyFormatter('%(asctime)s %(levelname)-3.3s %(message)s', '%Y-%m-%d %H:%M:%S', encoding='utf-8')

    manager.sanitized_config_file_content=sanitized_config_file_content
    
    manager.attachments=[ ('config.ini', None, manager.sanitized_config_file_content.encode(cmd_options.encoding), 'text', 'plain', cmd_options.encoding), ] 
    
    for job_name in job_list:
    
        manager.start_logging()
    
        now=datetime.now()
        manager.now=now
        start=int(time.time())
        job=dict(config.items(job_name))
        program=job.get('program')
        arch=gen_archiver(program)()
        arch.__version__=__version__
        arch.night_shift=boolean.get(job.get('night_shift', 'on').lower(), True)
        job=dict(config.items(job_name))
        job['name']=job_name
        job['config']=cmd_options.config
        job['config_encoding']=cmd_options.encoding
                
        if job['up2date']=='no':
            job['msg_header']='\n%s\n' % job['up2date_msg']
        else:
            job['msg_header']=''
        
        mail_config_errors=check_mail_config(job)
    
        # check job parameters 
        
        try:
            errors, warnings, extra=arch.load(job, manager)
        except Exception, e:
            try:
                log.exception('loading job %s by program %s', job_name, program)
            except UnicodeDecodeError:
                log.error('loading job %s by program %s: %s', job_name, program, ErrorDecode(e, default_encoding))
                
            if job['mail'] and not mail_config_errors:
                subject='MKSBACKUP %s ERR %s ' % (command_display, job['name'], )
                mail_exception(start, job, '', manager, subject)
            continue
    
        for line in extra.split('\n'):
            if line:
                log.info('    %s', line)
    
        errors.update(mail_config_errors)
        
        msg_body=job['msg_header']
            
        if errors:
            exit_code=3
            msg_body+='Errors in section: %s\n\n' % job['name']
            log.error('Errors in section: %s', job['name'])
            for k, v in errors.iteritems():
                log.error('\tparameter "%s" : %s', k, v)
                msg_body+='       parameter "%s" : %s\n\n' % (k, v)
    
            for k, v in warnings.iteritems():
                log.warning('%s: %s', k, v)
                msg_body+='%s\n    %s\n\n' % (k, v)
                    
            if job['mail'] and not mail_config_errors and not command=='check':
                subject='MKSBACKUP %s ERR %s ' % (command_display, job['name'], )
                status=gen_status(start, job, 'ERR', None)
                attachments=manager.attachments[:]
                attachments.append(('status.txt', None, status.encode('utf-8'), 'text', 'plain', 'utf-8')),
                errmsg=sendmail(job, subject, msg_body, attachments, manager.log)
    
            continue
    
        if warnings:
            log.warning('Warnings in section: %s', job['name'])
            for k, v in warnings.iteritems():
                log.warning('%s: %s', k, v)
        else:
            log.info('No errors in section: %s', job['name'])
    
        destination=getattr(arch, 'destination', None)
        
        if command in ('check', 'checkmail'):
            msg_body=job['msg_header']
            if destination and isinstance(destination, Destinations): 
                msg_body+='Destinations day by day:\n\n'
                log.info('Destinations day by day:')
                for i in range(destination.full_cycle):
                    today=now+timedelta(days=i)
                    typ, target=destination.match(today, night_shift=arch.night_shift)
                    if typ=='none':
                        ty='none'
                        target='-----------------'
                    elif typ==None:
                        ty=''
                        target='??? no target ???'
                    else:
                        ty=typ
                    msg_body+='    %s %-12s %s\n' % (today.strftime('%a %d %b %Y'), ty, target)
                    log.info('    %s %-12s %s', today.strftime('%a %d %b %Y'), ty, target)

                    if typ!='none' and typ!=None:
                        if target.startswith('ftp://'):
                            msg_body+='      ftp directory not tested\n'
                            log.info('       ftp directory not tested: %s', target)
                        elif not os.path.isdir(os.path.dirname(target)):
                            msg_body+='        directory not found !!!\n'
                            log.warning('        directory not found: %s', os.path.dirname(target))
                        
            msg_body='%s\n\n%s\n' % (msg_body, extra)
            if job['mail'] and command=='checkmail':
                subject='MKSBACKUP %s OK %s ' % (command_display, job['name'], )
                errmsg=sendmail(job, subject, msg_body, manager.attachments, manager.log)
            continue 
    
        #print 'destination', destination
        if destination and isinstance(destination, Destinations): 
            typ, target=destination.match(now, night_shift=arch.night_shift)
            if typ=='none' or typ==None:
                log.info('bye, no backup today')
                continue
    
        log.info('start command=%s job=%s archiver=%s', command, job_name, program)
        try:
            result=arch.run(command, job, manager)
        except Exception, e:
            exit_code=2
            try:
                log.exception('running command=%s job=%s archiver=%s', command, job_name, program)
            except UnicodeDecodeError:
                log.error('running command=%s job=%s archiver=%s : %s', command, job_name, program, ErrorDecode(e, default_encoding))

            if job['mail']:
                subject='MKSBACKUP BACKUP ERR %s ' % (job['name'], )
                mail_exception(start, job, extra, manager, subject)
        else:
            log.info('end command=%s job=%s archiver=%s', command, job_name, program)
    
            if isinstance(result, int):
                exit_code=max(exit_code, result)
            else:
                if result[0]!='OK':
                    exit_code=max(exit_code, 1)
    
                errmsg=send_mail_report(job, result, manager)

        
    log.info('final exit code=%d', exit_code)

    return exit_code
    
if __name__ == "__main__":
    sys.exit(main())


#<full,first su>S:\monthly\data-full-month-${month%2}.bkf
#<full,su>S:\week${week%4}\data-full-%a.bkf
#<inc,mo-sa>S:\week${week%4}\data-inc-%a.bkf



#<full,week%4=0>S:\data-full.bkf
#<inc,week%4=1>S:\data-inc.bkf

#
#w1/4
#jan/2,su
#feb-dec/2


# when start a week
# firstday=sat
#last day OF THE MONTH
#1st mon,2nd tue,3th wed,4th thu,last sat  OF THE MONTH
#1st week,2nd week,3th week,4th week,last week OF THE MONTH
#<> => empty selector to allow "no backup"
# empty target => no backup



# http://www.scalabium.com/faq/dct0082.htm
# To detect the code page of Windows operation system you must call the GetACP function from Windows API.
# If you needs to read the code page of "DOS" sessions, you must call the GetOEMCP function from Windows API. This function will return the value:
# http://msdn.microsoft.com/en-us/library/dd318070%28VS.85%29.aspx

#


# windows unicode file http://www.helpware.net/FAR/help/Unicode2.htm
# Universal Encoding Detector http://chardet.feedparser.org/docs/supported-encodings.html


# wbadmin
#
# How to use wbadmin to backup in Windows Vista 2008
# http://www.howtonetworking.com/Windows/wbadmin1.htm
#
# Backup Version and Space Management in Windows Server Backup
# https://blogs.technet.com/filecab/archive/2009/06/22/backup-version-and-space-management-in-windows-server-backup.aspx
#
# Customizing Windows Server Backup Schedule
# http://blogs.technet.com/filecab/archive/2009/04/13/customizing-windows-server-backup-schedule.aspx
#
# %windir%\logs\windowsserverbackup
#
# Output of command "Vshadow -wm2" ran on an elevated command prompt
#
# Deleting System State Backup:
# wbadmin delete systemstatebackup -deleteOldest
# wbadmin delete systemstatebackup -keepversions:10   
# The above command will keep the latest 10 versions and delete the rest all the system state backups.
#
# Deleting other backups:  Backup application stores multiple backup versions in the VSS shadow copies. Hence, older backup version can be deleted by deleting older shadow copy. Commands to list and delete VSS shadow copies are below. They need to be run in an elevated command window.
# vssadmin list shadows /for=x: ? for listing the snapshots on x: where x: is the backup location
# vssadmin delete shadows /for=x:  /oldest ? for delete the oldest shadow copy. It can be called multiple times in case there is need to delete multiple older backups.
# Note: wbadmin get versions or backup UI would still report deleted backups until next backup runs. At end of each backup ? non-existent backups are removed from the backup catalog.
#
# Win2k8: the backups are stored inside the    <drive_letter>:\WindowsImageBackup\<machineName>\SystemStateBackup
# Win2k8 R2: the backups are stored inside the <drive_letter>:\WindowsImageBackup\<machineName>\BackupSet<..>\ path




#
# (C) Copyright 2006, 2007 Hewlett-Packard Development Company, L.P.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Tim Potter <tpot@hp.com>

from xml.dom.pulldom import *
from pywbem import *

class ParseError(Exception):
    """This exception is raised when there is a validation error detected
    by the parser."""    
    pass

#
# Helper functions
#

def get_required_attribute(node, attr):
    """Return an attribute by name.  Throw an exception if not present."""
    
    if not node.hasAttribute(attr):
        raise ParseError(
            'Expecting %s attribute in element %s' % (attr, node.tagName))

    return node.getAttribute(attr)

def get_attribute(node, attr):
    """Return an attribute by name, or None if not present."""
    
    if node.hasAttribute(attr):
        return node.getAttribute(attr)

    return None

def get_end_event(parser, tagName):
    """Check that the next event is the end of a particular tag."""

    (event, node) = parser.next()

    if event != END_ELEMENT or node.tagName != tagName:
        raise ParseError(
            'Expecting %s end tag, got %s %s' % (tagName, event, node.tagName))

def is_start(event, node, tagName):
    """Return true if (event, node) is a start event for tagname."""
    return event == START_ELEMENT and node.tagName == tagName

def is_end(event, node, tagName):
    """Return true if (event, node) is an end event for tagname."""
    return event == END_ELEMENT and node.tagName == tagName

# <!-- ************************************************** -->
# <!-- Root element                                       -->
# <!-- ************************************************** -->

# <!ELEMENT CIM (MESSAGE | DECLARATION)>
# <!ATTLIST CIM
# 	CIMVERSION CDATA #REQUIRED
# 	DTDVERSION CDATA #REQUIRED
# >

# <!-- ************************************************** -->
# <!-- Object declaration elements                        -->
# <!-- ************************************************** -->

# <!ELEMENT DECLARATION (DECLGROUP | DECLGROUP.WITHNAME | DECLGROUP.WITHPATH)+>
# <!ELEMENT DECLGROUP ((LOCALNAMESPACEPATH | NAMESPACEPATH)?, QUALIFIER.DECLARATION*, VALUE.OBJECT*)>
# <!ELEMENT DECLGROUP.WITHNAME ((LOCALNAMESPACEPATH | NAMESPACEPATH)?, QUALIFIER.DECLARATION*, VALUE.NAMEDOBJECT*)>
# <!ELEMENT DECLGROUP.WITHPATH (VALUE.OBJECTWITHPATH | VALUE.OBJECTWITHLOCALPATH)*>
# <!ELEMENT QUALIFIER.DECLARATION (SCOPE?, (VALUE | VALUE.ARRAY)?)>
# <!ATTLIST QUALIFIER.DECLARATION 
#          %CIMName;               
#          %CIMType;               #REQUIRED
#          ISARRAY    (true|false) #IMPLIED
#          %ArraySize;
#          %QualifierFlavor;>
# <!ELEMENT SCOPE EMPTY>
# <!ATTLIST SCOPE
# 	CLASS (true | false) "false"
# 	ASSOCIATION (true | false) "false"
# 	REFERENCE (true | false) "false"
# 	PROPERTY (true | false) "false"
# 	METHOD (true | false) "false"
# 	PARAMETER (true | false) "false"
# 	INDICATION (true | false) "false"
# >

# <!-- ************************************************** -->
# <!-- Object Value elements                              -->
# <!-- ************************************************** -->

# <!ELEMENT VALUE (#PCDATA)>

def parse_value(parser, event, node):

    value = ''

    (next_event, next_node) = parser.next()

    if next_event == CHARACTERS:        

        value = next_node.nodeValue

        (next_event, next_node) = parser.next()
        
    if not is_end(next_event, next_node, 'VALUE'):
        raise ParseError('Expecting end VALUE')
        
    return value

# <!ELEMENT VALUE.ARRAY (VALUE*)>

def parse_value_array(parser, event, node):

    value_array = []

    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'VALUE'):

        value_array.append(parse_value(parser, next_event, next_node))

        while 1:

            (next_event, next_node) = parser.next()

            if is_end(next_event, next_node, 'VALUE.ARRAY'):
                break

            if is_start(next_event, next_node, 'VALUE'):
                value_array.append(parse_value(parser, next_event, next_node))
            else:
                raise ParseError('Expecting VALUE element')

    return value_array

# <!ELEMENT VALUE.REFERENCE (CLASSPATH | LOCALCLASSPATH | CLASSNAME |
#                            INSTANCEPATH | LOCALINSTANCEPATH | INSTANCENAME)>

def parse_value_reference(parser, event, node):

    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'CLASSPATH'):
        result = parse_classpath(parser, next_event, next_node)

    elif is_start(next_event, next_node, 'LOCALCLASSPATH'):
        result = parse_localclasspath(parser, next_event, next_node)

    elif is_start(next_event, next_node, 'CLASSNAME'):
        result = parse_classname(parser, next_event, next_node)

    elif is_start(next_event, next_node, 'INSTANCEPATH'):
        result = parse_instancepath(parser, next_event, next_node)

    elif is_start(next_event, next_node, 'LOCALINSTANCEPATH'):
        result = parse_localinstancepath(parser, next_event, next_node)

    elif is_start(next_event, next_node, 'INSTANCENAME'):
        result = parse_instancename(parser, next_event, next_node)
        
    else:
        raise ParseError('Expecting (CLASSPATH | LOCALCLASSPATH | CLASSNAME '
                         '| INSTANCEPATH | LOCALINSTANCEPATH | INSTANCENAME)')
                         
    get_end_event(parser, 'VALUE.REFERENCE')

    return result

# <!ELEMENT VALUE.REFARRAY (VALUE.REFERENCE*)>
# <!ELEMENT VALUE.OBJECT (CLASS | INSTANCE)>
# <!ELEMENT VALUE.NAMEDINSTANCE (INSTANCENAME, INSTANCE)>
# <!ELEMENT VALUE.NAMEDOBJECT (CLASS | (INSTANCENAME, INSTANCE))>
# <!ELEMENT VALUE.OBJECTWITHLOCALPATH ((LOCALCLASSPATH, CLASS) | (LOCALINSTANCEPATH, INSTANCE))>
# <!ELEMENT VALUE.OBJECTWITHPATH ((CLASSPATH, CLASS) | (INSTANCEPATH, INSTANCE))>
# <!ELEMENT VALUE.NULL EMPTY>
 
# <!-- ************************************************** -->
# <!-- Object naming and locating elements                -->
# <!-- ************************************************** -->

# <!ELEMENT NAMESPACEPATH (HOST, LOCALNAMESPACEPATH)>

def parse_namespacepath(parser, event, node):
    
    (next_event, next_node) = parser.next()

    if not is_start(next_event, next_node, 'HOST'):
        raise ParseError('Expecting HOST')
    
    host = parse_host(parser, next_event, next_node)

    (next_event, next_node) = parser.next()

    if not is_start(next_event, next_node, 'LOCALNAMESPACEPATH'):
        raise ParseError('Expecting LOCALNAMESPACEPATH')
    
    namespacepath = parse_localnamespacepath(parser, next_event, next_node)

    (next_event, next_node) = parser.next()

    if not is_end(next_event, next_node, 'NAMESPACEPATH'):
        raise ParseError('Expecting end NAMESPACEPATH')

    return (host, namespacepath)

# <!ELEMENT LOCALNAMESPACEPATH (NAMESPACE+)>

def parse_localnamespacepath(parser, event, node):

    (next_event, next_node) = parser.next()

    namespaces = []

    if not is_start(next_event, next_node, 'NAMESPACE'):
        print next_event, next_node
        raise ParseError('Expecting NAMESPACE')

    namespaces.append(parse_namespace(parser, next_event, next_node))

    while 1:

        (next_event, next_node) = parser.next()

        if is_end(next_event, next_node, 'LOCALNAMESPACEPATH'):
            break

        if is_start(next_event, next_node, 'NAMESPACE'):
            namespaces.append(parse_namespace(parser, next_event, next_node))
        else:
            raise ParseError('Expecting NAMESPACE')

    return string.join(namespaces, '/')

# <!ELEMENT HOST (#PCDATA)>

def parse_host(parser, event, node):

    host = ''

    (next_event, next_node) = parser.next()

    if next_event == CHARACTERS:        

        host = next_node.nodeValue

        (next_event, next_node) = parser.next()
        
    if not is_end(next_event, next_node, 'HOST'):
        raise ParseError('Expecting end HOST')
        
    return host

# <!ELEMENT NAMESPACE EMPTY>
# <!ATTLIST NAMESPACE
# 	%CIMName; 
# >

def parse_namespace(parser, event, node):

    name = get_required_attribute(node, 'NAME')

    (next_event, next_node) = parser.next()

    if not is_end(next_event, next_node, 'NAMESPACE'):
        raise ParseError('Expecting end NAMESPACE')

    return name

# <!ELEMENT CLASSPATH (NAMESPACEPATH, CLASSNAME)>

# <!ELEMENT LOCALCLASSPATH (LOCALNAMESPACEPATH, CLASSNAME)>

# <!ELEMENT CLASSNAME EMPTY>
# <!ATTLIST CLASSNAME
# 	%CIMName; 
# >

# <!ELEMENT INSTANCEPATH (NAMESPACEPATH, INSTANCENAME)>

def parse_instancepath(parser, event, node):

    (next_event, next_node) = parser.next()

    if not is_start(next_event, next_node, 'NAMESPACEPATH'):
        raise ParseError('Expecting NAMESPACEPATH')
    
    host, namespacepath = parse_namespacepath(parser, next_event, next_node)

    (next_event, next_node) = parser.next()

    if not is_start(next_event, next_node, 'INSTANCENAME'):
        print next_event, next_node
        raise ParseError('Expecting INSTANCENAME')

    instancename = parse_instancename(parser, next_event, next_node)

    instancename.host = host
    instancename.namespace = namespacepath

    return instancename

# <!ELEMENT LOCALINSTANCEPATH (LOCALNAMESPACEPATH, INSTANCENAME)>

def parse_localinstancepath(parser, event, node):

    (next_event, next_node) = parser.next()

    if not is_start(next_event, next_node, 'LOCALNAMESPACEPATH'):
        raise ParseError('Expecting LOCALNAMESPACEPATH')
    
    namespacepath = parse_localnamespacepath(parser, next_event, next_node)

    (next_event, next_node) = parser.next()

    if not is_start(next_event, next_node, 'INSTANCENAME'):
        raise ParseError('Expecting INSTANCENAME')

    instancename = parse_instancename(parser, next_event, next_node)

    instancename.namespace = namespacepath

    return instancename

# <!ELEMENT INSTANCENAME (KEYBINDING* | KEYVALUE? | VALUE.REFERENCE?)>
# <!ATTLIST INSTANCENAME
# 	%ClassName; 
# >

def parse_instancename(parser, event, node):
    
    classname = get_required_attribute(node, 'CLASSNAME')
    keybindings = []
    
    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'KEYBINDING'):

        keybindings.append(parse_keybinding(parser, next_event, next_node))

        while 1:

            (next_event, next_node) = parser.next()

            if is_end(next_event, next_node, 'INSTANCENAME'):
                break

            if is_start(next_event, next_node, 'KEYBINDING'):
                keybindings.append(
                    parse_keybinding(parser, next_event, next_node))
            else:
                raise ParseError('Expecting KEYBINDING element')
            
    if is_end(next_event, next_node, 'INSTANCENAME'):
        pass
            
    elif is_start(next_event, next_node, 'KEYVALUE'):
        keybindings.append(('', parse_keyvalue(parser, next_event, next_node)))
        
    elif is_start(next_event, next_node, 'VALUE.REFERENCE'):        
        keybindings.append(
            parse_value_reference(parser, next_event, next_node))
        
    else:
        raise ParseError(
            'Expecting KEYBINDING* | KEYVALUE? | VALUE.REFERENCE')

    return CIMInstanceName(classname, keybindings)

# <!ELEMENT OBJECTPATH (INSTANCEPATH | CLASSPATH)>

# <!ELEMENT KEYBINDING (KEYVALUE | VALUE.REFERENCE)>
# <!ATTLIST KEYBINDING
# 	%CIMName; 
# >

def parse_keybinding(parser, event, node):

    name = get_required_attribute(node, 'NAME')

    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'KEYVALUE'):
        keyvalue = parse_keyvalue(parser, next_event, next_node)
        result = (name, keyvalue)

    elif is_start(next_event, next_node, 'VALUE.REFERENCE'):
        value_reference = parse_value_reference(parser, next_event, next_node)
        result = (name, value_reference)

    else:
        raise ParseError('Expecting KEYVALUE or VALUE.REFERENCE element')

    get_end_event(parser, 'KEYBINDING')

    return result

# <!ELEMENT KEYVALUE (#PCDATA)>
# <!ATTLIST KEYVALUE
# 	VALUETYPE (string | boolean | numeric) "string"
#         %CIMType;              #IMPLIED>

def parse_keyvalue(parser, event, node):

    valuetype = get_required_attribute(node, 'VALUETYPE')
    type = get_attribute(node, 'TYPE')

    (next_event, next_node) = parser.next()

    if next_event != CHARACTERS:
        raise ParseError('Expecting character data')
        
    value = next_node.nodeValue

    if valuetype == 'string':
        pass

    elif valuetype == 'boolean':

        # CIM-XML says "These values MUST be treated as
        # case-insensitive" (even though the XML definition
        # requires them to be lowercase.)

        p = value.strip().lower()

        if p == 'true':
            value = True
        elif p == 'false':
            value = False
        else:
            raise ParseError('invalid boolean value "%s"' % `p`)

    elif valuetype == 'numeric':

        try: 

            # XXX: Use TYPE attribute to create named CIM type.
            # if attrs(tt).has_key('TYPE'):
            #    return cim_obj.tocimobj(attrs(tt)['TYPE'], p.strip())

            # XXX: Would like to use long() here, but that tends to cause
            # trouble when it's written back out as '2L'

            value = int(value.strip(), 0)

        except ValueError:
            raise ParseError(
                'invalid numeric value "%s"' % value)

    else:
        raise ParseError('Invalid VALUETYPE')

    get_end_event(parser, 'KEYVALUE')

    return value

# <!-- ************************************************** -->
# <!-- Object definition elements                         -->
# <!-- ************************************************** -->

# <!ELEMENT CLASS (QUALIFIER*, (PROPERTY | PROPERTY.ARRAY | PROPERTY.REFERENCE)*, METHOD*)>
# <!ATTLIST CLASS
# 	%CIMName; 
# 	%SuperClass; 
# >

# <!ELEMENT INSTANCE (QUALIFIER*, (PROPERTY | PROPERTY.ARRAY |
#                     PROPERTY.REFERENCE)*)>
# <!ATTLIST INSTANCE
# 	%ClassName;
#         xml:lang   NMTOKEN      #IMPLIED 
# >

def parse_instance(parser, event, node):

    classname = get_required_attribute(node, 'CLASSNAME')

    properties = []
    qualifiers = []

    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'QUALIFIER'):

        qualifiers.append(parse_qualifier(parser, next_event, next_node))

        while 1:

            (next_event, next_node) = parser.next()

            if is_start(next_event, next_node, 'PROPERTY') or \
               is_start(next_event, next_node, 'PROPERTY.ARRAY') or \
               is_start(next_event, next_node, 'PROPERTY.REFERENCE') or \
               is_end(next_event, next_node, 'INSTANCE'):
                break
        
            if is_start(next_event, next_node, 'QUALIFIER'):
                qualifiers.append(
                    parse_qualifier(parser, next_event, next_node))
            else:
                raise ParseError('Expecting QUALIFIER')

    while 1:

        if is_end(next_event, next_node, 'INSTANCE'):
            break

        if is_start(next_event, next_node, 'PROPERTY'):
            properties.append(parse_property(parser, next_event, next_node))

        elif is_start(next_event, next_node, 'PROPERTY.ARRAY'):
            properties.append(
                parse_property_array(parser, next_event, next_node))

        elif is_start(next_event, next_node, 'PROPERTY.REFERENCE'):
            properties.append(
                parse_property_reference(parser, next_event, next_node))
        else:
            raise ParseError(
                'Expecting (PROPERTY | PROPERTY.ARRAY | PROPERTY.REFERENCE)')

        (next_event, next_node) = parser.next()

    if not is_end(next_event, next_node, 'INSTANCE'):
        raise ParseError('Expecting end INSTANCE')

    return CIMInstance(classname,
                       properties = dict([(x.name,  x) for x in properties]),
                       qualifiers = dict([(x.name, x) for x in qualifiers]))
                       
        
# <!ELEMENT QUALIFIER ((VALUE | VALUE.ARRAY)?)>
# <!ATTLIST QUALIFIER 
#          %CIMName;
#          %CIMType;              #REQUIRED
#          %Propagated;
#          %QualifierFlavor;
#          xml:lang   NMTOKEN     #IMPLIED 
# >

def parse_qualifier(parser, event, node):

    name = get_required_attribute(node, 'NAME')
    type = get_required_attribute(node, 'TYPE')
    propagated = get_attribute(node, 'PROPAGATED')

    (next_event, next_node) = parser.next()

    if is_end(next_event, next_node, 'QUALIFIER'):
        return CIMQualifier(name, None, type = type)

    if is_start(next_event, next_node, 'VALUE'):
        value = parse_value(parser, next_event, next_node)
    elif is_start(next_event, next_node, 'VALUE.ARRAY'):
        value = parse_value_array(parser, next_event, next_node)
    else:
        raise ParseError('Expecting (VALUE | VALUE.ARRAY)')

    result = CIMQualifier(name, tocimobj(type, value))

    get_end_event(parser, 'QUALIFIER')

    return result

# <!ELEMENT PROPERTY (QUALIFIER*, VALUE?)>
# <!ATTLIST PROPERTY 
#          %CIMName;
#          %ClassOrigin;
#          %Propagated;
#          %CIMType;              #REQUIRED
#          xml:lang   NMTOKEN     #IMPLIED 
# >

def parse_property(parser, event, node):
 
    name = get_required_attribute(node, 'NAME')
    type = get_required_attribute(node, 'TYPE')

    class_origin = get_attribute(node, 'CLASSORIGIN')
    propagated = get_attribute(node, 'PROPAGATED')

    qualifiers = []
    value = None

    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'QUALIFIER'):

        qualifiers.append(parse_qualifier(parser, next_event, next_node))

        while 1:

            (next_event, next_node) = parser.next()

            if is_start(next_event, next_node, 'VALUE'):
                break

            if is_start(next_event, next_node, 'QUALIFIER'):
                qualifiers.append(
                    parse_qualifier(parser, next_event, next_node))
            else:
                raise ParseError('Expecting QUALIFIER')

    if is_start(next_event, next_node, 'VALUE'):

        value = parse_value(parser, next_event, next_node)
        (next_event, next_node) = parser.next()

    if not is_end(next_event, next_node, 'PROPERTY'):
        raise ParseError('Expecting end PROPERTY')

    return CIMProperty(name,
                       tocimobj(type, value),
                       type = type,
                       class_origin = class_origin,
                       propagated = propagated,
                       qualifiers = dict([(x.name, x) for x in qualifiers]))

# <!ELEMENT PROPERTY.ARRAY (QUALIFIER*, VALUE.ARRAY?)>
# <!ATTLIST PROPERTY.ARRAY 
#          %CIMName;
#          %CIMType;              #REQUIRED
#          %ArraySize;
#          %ClassOrigin;
#          %Propagated;
#          xml:lang   NMTOKEN     #IMPLIED 
# >

def parse_property_array(parser, event, node):
 
    name = get_required_attribute(node, 'NAME')
    type = get_required_attribute(node, 'TYPE')

    array_size = get_attribute(node, 'ARRAYSIZE')
    class_origin = get_attribute(node, 'CLASSORIGIN')
    propagated = get_attribute(node, 'PROPAGATED')

    qualifiers = []
    value = None

    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'QUALIFIER'):

        qualifiers.append(parse_qualifier(parser, next_event, next_node))

        while 1:

            (next_event, next_node) = parser.next()

            if is_start(next_event, next_node, 'VALUE.ARRAY'):
                break

            if is_start(next_event, next_node, 'QUALIFIER'):
                qualifiers.append(
                    parse_qualifier(parser, next_event, next_node))
            else:
                raise ParseError('Expecting QUALIFIER')

    if is_start(next_event, next_node, 'VALUE.ARRAY'):

        value = parse_value_array(parser, next_event, next_node)
        (next_event, next_node) = parser.next()

    if not is_end(next_event, next_node, 'PROPERTY.ARRAY'):
        raise ParseError('Expecting end PROPERTY.ARRAY')

    return CIMProperty(name,
                       tocimobj(type, value),
                       type = type,
                       class_origin = class_origin,
                       propagated = propagated,
                       is_array = True,
                       qualifiers = dict([(x.name, x) for x in qualifiers]))
    
# <!ELEMENT PROPERTY.REFERENCE (QUALIFIER*, (VALUE.REFERENCE)?)>
# <!ATTLIST PROPERTY.REFERENCE
# 	%CIMName; 
# 	%ReferenceClass; 
# 	%ClassOrigin; 
# 	%Propagated; 
# >

def parse_property_reference(parser, event, node):

    name = get_required_attribute(node, 'NAME')

    class_origin = get_attribute(node, 'CLASSORIGIN')
    propagated = get_attribute(node, 'PROPAGATED')

    qualifiers = []
    value = None

    (next_event, next_node) = parser.next()

    if is_start(next_event, next_node, 'QUALIFIER'):

        qualifiers.append(parse_qualifier(parser, next_event, next_node))

        while 1:

            (next_event, next_node) = parser.next()

            if is_start(next_event, next_node, 'VALUE.REFERENCE'):
                break

            if is_start(next_event, next_node, 'QUALIFIER'):
                qualifiers.append(
                    parse_qualifier(parser, next_event, next_node))

            else:
                raise ParseError('Expecting QUALIFIER')

    if is_start(next_event, next_node, 'VALUE.REFERENCE'):

        value = parse_value_reference(parser, next_event, next_node)
        (next_event, next_node) = parser.next()

    if not is_end(next_event, next_node, 'PROPERTY.REFERENCE'):
        raise ParseError('Expecting end PROPERTY.REFERENCE')

    return CIMProperty(name,
                       value,
                       class_origin = class_origin,
                       propagated = propagated,
                       type = 'reference',
                       qualifiers = dict([(x.name, x) for x in qualifiers]))

# <!ELEMENT METHOD (QUALIFIER*, (PARAMETER | PARAMETER.REFERENCE | PARAMETER.ARRAY | PARAMETER.REFARRAY)*)>
# <!ATTLIST METHOD 
#          %CIMName;
#          %CIMType;              #IMPLIED
#          %ClassOrigin;
#          %Propagated;>
# <!ELEMENT PARAMETER (QUALIFIER*)>
# <!ATTLIST PARAMETER 
#          %CIMName;
#          %CIMType;              #REQUIRED>

# <!ELEMENT PARAMETER.REFERENCE (QUALIFIER*)>
# <!ATTLIST PARAMETER.REFERENCE
# 	%CIMName; 
# 	%ReferenceClass; 
# >

# <!ELEMENT PARAMETER.ARRAY (QUALIFIER*)>
# <!ATTLIST PARAMETER.ARRAY 
#          %CIMName;
#          %CIMType;              #REQUIRED
#          %ArraySize;>

# <!ELEMENT PARAMETER.REFARRAY (QUALIFIER*)>
# <!ATTLIST PARAMETER.REFARRAY
# 	%CIMName; 
# 	%ReferenceClass; 
# 	%ArraySize; 
# >

# <!ELEMENT TABLECELL.DECLARATION EMPTY> 
# <!ATTLIST TABLECELL.DECLARATION
#      %CIMName;
#      %CIMType;   	#REQUIRED
#      ISARRAY           (true|false) "false"
#      %ArraySize;
#      CELLPOS  		CDATA       #REQUIRED
#      SORTPOS     	CDATA       #IMPLIED
#      SORTDIR            (ASC|DESC)  #IMPLIED
# >

# <!ELEMENT TABLECELL.REFERENCE EMPTY> 
# <!ATTLIST TABLECELL.REFERENCE
#      %CIMName;
#      %ReferenceClass;	
#      ISARRAY        (true|false) "false"
#      %ArraySize; 
#      CELLPOS          CDATA       #REQUIRED
#      SORTPOS          CDATA       #IMPLIED
#      SORTDIR         (ASC|DESC)   #IMPLIED
# >

# <!ELEMENT TABLEROW.DECLARATION ( TABLECELL.DECLARATION | TABLECELL.REFERENCE)*>

# <!ELEMENT TABLE (TABLEROW.DECLARATION,(TABLEROW)*)>

# <!ELEMENT TABLEROW ( VALUE | VALUE.ARRAY | VALUE.REFERENCE | VALUE.REFARRAY | VALUE.NULL)*>
 
# <!-- ************************************************** -->
# <!-- Message elements                                   -->
# <!-- ************************************************** -->

# <!ELEMENT MESSAGE (SIMPLEREQ | MULTIREQ | SIMPLERSP | MULTIRSP | SIMPLEEXPREQ | MULTIEXPREQ | SIMPLEEXPRSP | MULTIEXPRSP)>
# <!ATTLIST MESSAGE
# 	ID CDATA #REQUIRED
# 	PROTOCOLVERSION CDATA #REQUIRED
# >

# <!ELEMENT MULTIREQ (SIMPLEREQ, SIMPLEREQ+)>

# <!ELEMENT MULTIEXPREQ (SIMPLEEXPREQ, SIMPLEEXPREQ+)>

# <!ELEMENT SIMPLEREQ (IMETHODCALL | METHODCALL)>

# <!ELEMENT SIMPLEEXPREQ (EXPMETHODCALL)>

# <!ELEMENT IMETHODCALL (LOCALNAMESPACEPATH, IPARAMVALUE*, RESPONSEDESTINATION?)>
# <!ATTLIST IMETHODCALL
# 	%CIMName; 
# >

# <!ELEMENT METHODCALL ((LOCALINSTANCEPATH | LOCALCLASSPATH), PARAMVALUE*, RESPONSEDESTINATION?)>
# <!ATTLIST METHODCALL
# 	%CIMName; 
# >

# <!ELEMENT EXPMETHODCALL (EXPPARAMVALUE*)>
# <!ATTLIST EXPMETHODCALL
# 	%CIMName; 
# >

# <!ELEMENT PARAMVALUE (VALUE | VALUE.REFERENCE | VALUE.ARRAY | VALUE.REFARRAY)?>
# <!ATTLIST PARAMVALUE
# 	%CIMName; 
#         %ParamType;  #IMPLIED
# >

# <!ELEMENT IPARAMVALUE (VALUE | VALUE.ARRAY | VALUE.REFERENCE | INSTANCENAME | CLASSNAME | QUALIFIER.DECLARATION | CLASS | INSTANCE | VALUE.NAMEDINSTANCE)?>
# <!ATTLIST IPARAMVALUE
# 	%CIMName; 
# >

# <!ELEMENT EXPPARAMVALUE (INSTANCE?|VALUE?|METHODRESPONSE?|IMETHODRESPONSE?)>
# <!ATTLIST EXPPARAMVALUE
# 	%CIMName; 
# >

# <!ELEMENT MULTIRSP (SIMPLERSP, SIMPLERSP+)>

# <!ELEMENT MULTIEXPRSP (SIMPLEEXPRSP, SIMPLEEXPRSP+)>

# <!ELEMENT SIMPLERSP (METHODRESPONSE | IMETHODRESPONSE | SIMPLEREQACK)>

# <!ELEMENT SIMPLEEXPRSP (EXPMETHODRESPONSE)>

# <!ELEMENT METHODRESPONSE (ERROR|(RETURNVALUE?,PARAMVALUE*))>
# <!ATTLIST METHODRESPONSE
# 	%CIMName; 
# >

# <!ELEMENT EXPMETHODRESPONSE (ERROR|IRETURNVALUE?)>
# <!ATTLIST EXPMETHODRESPONSE
# 	%CIMName; 
# >

# <!ELEMENT IMETHODRESPONSE (ERROR|IRETURNVALUE?)>
# <!ATTLIST IMETHODRESPONSE
# 	%CIMName; 
# >

# <!ELEMENT ERROR (INSTANCE*)>
# <!ATTLIST ERROR
# 	CODE CDATA #REQUIRED
# 	DESCRIPTION CDATA #IMPLIED
# >

# <!ELEMENT RETURNVALUE (VALUE | VALUE.REFERENCE)>
# <!ATTLIST RETURNVALUE
# 	%ParamType;       #IMPLIED 
# >

# <!ELEMENT IRETURNVALUE (CLASSNAME* | INSTANCENAME* | VALUE* | VALUE.OBJECTWITHPATH* | VALUE.OBJECTWITHLOCALPATH* | VALUE.OBJECT* | OBJECTPATH* | QUALIFIER.DECLARATION* | VALUE.ARRAY? | VALUE.REFERENCE? | CLASS* | INSTANCE* | VALUE.NAMEDINSTANCE*)>

# <!ELEMENT RESPONSEDESTINATION (INSTANCE)> 

# <!ELEMENT SIMPLEREQACK (ERROR?)> 
# <!ATTLIST SIMPLEREQACK  
#          INSTANCEID CDATA     #REQUIRED
# > 

def make_parser(stream_or_string):
    """Create a xml.dom.pulldom parser."""
    
    if type(stream_or_string) == str or type(stream_or_string) == unicode:

        # XXX: the pulldom.parseString() function doesn't seem to
        # like operating on unicode strings!
        
        return parseString(str(stream_or_string))
        
    else:

        return parse(stream_or_string)

def parse_any(stream_or_string):
    """Parse any XML string or stream."""
    
    parser = make_parser(stream_or_string)

    (event, node) = parser.next()

    if event != START_DOCUMENT:
        raise ParseError('Expecting document start')

    (event, node) = parser.next()

    if event != START_ELEMENT:
        raise ParseError('Expecting element start')

    fn_name = 'parse_%s' % node.tagName.lower().replace('.', '_')
    fn = globals().get(fn_name)
    if fn is None:
        raise ParseError('No parser for element %s' % node.tagName)
    
    return fn(parser, event, node)

# Test harness

if __name__ == '__main__':
    import sys
    print parse_any(sys.stdin)



#!/usr/bin/python

# (C) Copyright 2008 Hewlett-Packard Development Company, L.P.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# Author: Tim Potter <tpot@hp.com>

#
# A small utility to wrap up a PyWBEM session in a Python interactive
# console.
#
# Usage: 
#
#   wbemcli.py HOSTNAME [-u USERNAME -p PASSWORD] [-n namespace] [--no-ssl] \
#       [--port PORT]
#
# CIM operations can be executed by using the PyWBEM connection object
# called 'cli' in the global scope.  There are two sets of aliases
# available for usage in the interpreter. For example the following
# three commands are equivalent:
#
#   >>> cli.EnumerateInstanceNames('SMX_ComputerSystem')
#   >>> EnumerateInstanceNames('SMX_ComputerSystem')
#   >>> ein('SMX_ComputerSystem')
#
# Pretty-printing of results is also available using the 'pp'
# function. For example:
#
#   >>> cs = ei('SMX_ComputerSystem')[0]
#   >>> pp(cs.items())
#   [(u'RequestedState', 12L),
#    (u'Dedicated', [1L]),
#    (u'StatusDescriptions', [u'System is Functional']),
#    (u'IdentifyingNumber', u'6F880AA1-F4F5-11D5-8C45-C0116FBAE02A'),
#   ...
#

import os, stat, sys, string, getpass, errno
from pywbem import *
from code import InteractiveConsole
from optparse import OptionParser
from pprint import pprint as pp

# Conditional support of readline module

have_readline = False

try:
    import readline
    have_readline = True
except ImportError, arg:
    pass

#
# Parse command line args
#

optparser = OptionParser(
    usage = '%prog HOSTNAME [-u USER -p PASS] [-n NAMESPACE] [--no-ssl]')

# Username and password

optparser.add_option('-u', '--user', dest = 'user',
                     action = 'store', type = 'string',
                     help = 'user to connect as')

optparser.add_option('-p', '--password', dest = 'password',
                     action = 'store', type = 'string',
                     help = 'password to connect user as')

# Change the default namespace used

optparser.add_option('-n', '--namespace', dest = 'namespace',
                     action = 'store', type = 'string', default = 'root/cimv2',
                     help = 'default namespace to use')

# Don't use SSL for remote connections

optparser.add_option('--no-ssl', dest = 'no_ssl', action = 'store_true',
                     help = 'don\'t use SSL')

# Specify non-standard port

optparser.add_option('--port', dest = 'port', action = 'store', type = 'int',
                     help = 'port to connect as', default = None)

# Check usage

(opts, argv) = optparser.parse_args()

if len(argv) != 1:
    optparser.print_usage()
    sys.exit(1)

#
# Set up a client connection
#

def remote_connection():
    """Initiate a remote connection, via PyWBEM."""

    if argv[0][0] == '/':
        url = argv[0]
    else:
        proto = 'https'

        if opts.no_ssl:
            proto = 'http'

        url = '%s://%s' % (proto, argv[0])

        if opts.port is not None:
            url += ':%d' % opts.port

    creds = None

    if opts.user is not None and opts.password is None:
        opts.password = getpass.getpass('Enter password for %s: ' % opts.user)

    if opts.user is not None or opts.password is not None:
        creds = (opts.user, opts.password)

    cli = WBEMConnection(url, creds, default_namespace = opts.namespace)

    cli.debug = True

    return cli

cli = remote_connection()

#
# Create some convenient global functions to reduce typing
#

def EnumerateInstanceNames(classname, namespace = None):
    """Enumerate the names of the instances of a CIM Class (including the
    names of any subclasses) in the target namespace."""

    return cli.EnumerateInstanceNames(classname, namespace = namespace)

def EnumerateInstances(classname, namespace = None, LocalOnly = True,
                       DeepInheritance = True, IncludeQualifiers = False,
                       IncludeClassOrigin = False):
    """Enumerate instances of a CIM Class (includeing the instances of
    any subclasses in the target namespace."""

    return cli.EnumerateInstances(classname, 
                                  namespace = namespace,
                                  DeepInheritance = DeepInheritance,
                                  IncludeQualifiers = IncludeQualifiers,
                                  IncludeClassOrigin = IncludeClassOrigin)


def GetInstance(instancename, LocalOnly = True, IncludeQualifiers = False,
                IncludeClassOrigin = False):
    """Return a single CIM instance corresponding to the instance name
    given."""

    return cli.GetInstance(instancename, 
                           LocalOnly = LocalOnly, 
                           IncludeQualifiers = IncludeQualifiers,
                           IncludeClassOrigin = IncludeClassOrigin)

def DeleteInstance(instancename):
    """Delete a single CIM instance."""

    return cli.DeleteInstance(instancename)

def ModifyInstance(*args, **kwargs):
    return cli.ModifyInstance(*args, **kwargs)

def CreateInstance(*args, **kwargs):
    return cli.CreateInstance(*args, **kwargs)

def InvokeMethod(*args, **kwargs):
    return cli.InvokeMethod(*args, **kwargs)

def AssociatorNames(*args, **kwargs):
    return cli.AssociatorNames(*args, **kwargs)

def Associators(*args, **kwargs):
    return cli.Associators(*args, **kwargs)

def ReferenceNames(*args, **kwargs):
    return cli.ReferenceNames(*args, **kwargs)

def References(*args, **kwargs):
    return cli.References(*args, **kwargs)

def EnumerateClassNames(*args, **kwargs):
    return cli.EnumerateClassNames(*args, **kwargs)

def EnumerateClasses(*args, **kwargs):
    return cli.EnumerateClasses(*args, **kwargs)

def GetClass(*args, **kwargs):
    return cli.GetClass(*args, **kwargs)

def DeleteClass(*args, **kwargs):
    return cli.DeleteClass(*args, **kwargs)

def ModifyClass(*args, **kwargs):
    return cli.ModifyClass(*args, **kwargs)

def CreateClass(*args, **kwargs):
    return cli.CreateClass(*args, **kwargs)

def EnumerateQualifiers(*args, **kwargs):
    return cli.EnumerateQualifiers(*args, **kwargs)

def GetQualifier(*args, **kwargs):
    return cli.GetQualifier(*args, **kwargs)

def SetQualifier(*args, **kwargs):
    return cli.SetQualifier(*args, **kwargs)

def DeleteQualifier(*args, **kwargs):
    return cli.DeleteQualifier(*args, **kwargs)

# Aliases for global functions above

ein = EnumerateInstanceNames
ei = EnumerateInstances
gi = GetInstance
di = DeleteInstance
mi = ModifyInstance
ci = CreateInstance

im = InvokeMethod

an = AssociatorNames
ao = Associators
rn = ReferenceNames
re = References

ecn = EnumerateClassNames
ec = EnumerateClasses
gc = GetClass
dc = DeleteClass
mc = ModifyClass
cc = CreateClass

eq = EnumerateQualifiers
gq = GetQualifier
sq = SetQualifier
dq = DeleteQualifier

#
# Enter interactive console
#

def get_banner():
    
    result = ''

    # Note how we are connected

    result += 'Connected to %s' % cli.url
    if cli.creds is not None:
        result += ' as %s' % cli.creds[0]

    return result
        
# Read previous command line history

histfile = '%s/.wbemcli_history' % os.environ['HOME']

try:
    if have_readline:
        readline.read_history_file(histfile)
except IOError, arg:
    if arg[0] != errno.ENOENT:
        raise

# Interact

i = InteractiveConsole(globals())
i.interact(get_banner())

# Save command line history

if have_readline:
    readline.write_history_file(histfile)
